import pytest
import torch

from .stacker import Stacker, natural_sort, unstack


@pytest.mark.parametrize(
    "names, sorted_names",
    [
        (
            ["a_1", "b_1", "c_1", "a_2"],
            [
                "a_1",
                "a_2",
                "b_1",
                "c_1",
            ],
        ),
        (
            [
                "a_0",
                "a_1",
                "a_12",
                "a_2",
            ],
            [
                "a_0",
                "a_1",
                "a_2",
                "a_12",
            ],
        ),
        (
            [
                "a_0001",
                "a_0012",
                "a_0002",
            ],
            [
                "a_0001",
                "a_0002",
                "a_0012",
            ],
        ),
        (
            [
                "ab1",
                "aa10",
                "aa2",
            ],
            ["aa2", "aa10", "ab1"],
        ),
    ],
)
def test_natural_sort(names, sorted_names):
    assert natural_sort(names) == sorted_names


_PREFIX_MAPS = [{"a": ["aa_", "ab_"], "b": ["b_", "bb_"], "c": ["ca", "cb"]}]

_DATA_NAMES = [
    ["aa_0", "aa_1", "b_00", "b_01", "b", "ca"],
    ["ab_0", "ab_1", "bb_1", "bb_2", "cb"],
    ["ab_0", "b_", "cc"],
]


@pytest.mark.parametrize(
    "prefix_map, data_names, expected_level_names",
    [
        (
            _PREFIX_MAPS[0],
            _DATA_NAMES[0],
            {"a": ["aa_0", "aa_1"], "b": ["b_00", "b_01"], "c": ["ca"]},
        ),
        (
            _PREFIX_MAPS[0],
            _DATA_NAMES[1],
            {"a": ["ab_0", "ab_1"], "b": ValueError, "c": ["cb"]},
        ),
        (
            _PREFIX_MAPS[0],
            _DATA_NAMES[2],
            {"a": ["ab_0"], "b": ["b_"], "c": KeyError},
        ),
    ],
)
def test_get_all_level_names(prefix_map, data_names, expected_level_names):
    data = {name: torch.zeros(2, 2) for name in data_names}
    stacker = Stacker(prefix_map)
    for standard_name in expected_level_names.keys():
        if isinstance(expected_level_names[standard_name], type):
            with pytest.raises(expected_level_names[standard_name]):
                stacker.get_all_level_names(standard_name, data)
        else:
            assert (
                stacker.get_all_level_names(standard_name, data)
                == expected_level_names[standard_name]
            )


@pytest.mark.parametrize(
    "prefix_map, data_names, expected_level_names",
    [
        (
            _PREFIX_MAPS[0],
            _DATA_NAMES[0],
            {"a": ["aa_0", "aa_1"], "b": ["b_00", "b_01"], "c": ["ca"]},
        ),
        (
            _PREFIX_MAPS[0],
            _DATA_NAMES[1],
            {"a": ["ab_0", "ab_1"], "c": ["cb"]},
        ),
        (
            _PREFIX_MAPS[0],
            _DATA_NAMES[2],
            {"a": ["ab_0"], "b": ["b_"]},
        ),
    ],
)
def test_stack_unstack(prefix_map, data_names, expected_level_names):
    torch.manual_seed(0)
    data = {name: torch.rand(2, 2) for name in data_names}
    stacker = Stacker(prefix_map)
    for standard_name in expected_level_names.keys():
        level_names = expected_level_names[standard_name]
        if len(level_names) == 1:
            expected_stacked = data[level_names[0]].unsqueeze(-1)
        else:
            expected_stacked = torch.stack([data[name] for name in level_names], dim=-1)
        stacked = stacker(standard_name, data)
        assert torch.allclose(stacked, expected_stacked)
        unstacked = unstack(
            stacked, names=stacker.get_all_level_names(standard_name, data)
        )
        for name in level_names:
            assert name in unstacked
            assert torch.allclose(unstacked[name], data[name])



import dataclasses
import datetime
from typing import Any, Callable, Dict, List, Mapping, Optional, Tuple

import numpy as np
import torch
import xarray as xr
from matplotlib.figure import Figure

from fme.core.dataset.data_typing import VariableMetadata
from fme.core.device import get_device
from fme.core.distributed import Distributed
from fme.core.gridded_ops import GriddedOperations
from fme.core.typing_ import TensorMapping


class GlobalMeanAnnualAggregator:
    def __init__(
        self,
        ops: GriddedOperations,
        timestep: datetime.timedelta,
        variable_metadata: Optional[Mapping[str, VariableMetadata]] = None,
        monthly_reference_data: Optional[xr.Dataset] = None,
    ):
        self._area_weighted_mean = ops.area_weighted_mean
        self.timestep = timestep
        self.variable_metadata = variable_metadata
        self._target_datasets: Optional[List[xr.Dataset]] = None
        self._gen_datasets: Optional[List[xr.Dataset]] = None
        self._monthly_reference_data = monthly_reference_data
        self._variable_reference_data: Dict[str, VariableReferenceData] = {}

    def _get_reference(self, name: str) -> Optional["VariableReferenceData"]:
        if self._monthly_reference_data is None:
            return None
        if name not in self._variable_reference_data:
            if name not in self._monthly_reference_data:
                return None
            self._variable_reference_data[name] = process_monthly_reference(
                self._monthly_reference_data, self._area_weighted_mean, name
            )
        return self._variable_reference_data[name]

    @torch.no_grad()
    def record_batch(
        self,
        time: xr.DataArray,
        target_data: TensorMapping,
        gen_data: TensorMapping,
    ):
        """Record a batch of data for computing time variability statistics."""
        target_data_area_mean, gen_data_area_mean = {}, {}
        for name in gen_data.keys():
            target_data_area_mean[name] = self._area_weighted_mean(
                target_data[name]
            ).cpu()
            gen_data_area_mean[name] = self._area_weighted_mean(gen_data[name]).cpu()
        target_ds = to_dataset(target_data_area_mean, time)
        gen_ds = to_dataset(gen_data_area_mean, time)

        # must keep a separate dataset for each sample to avoid averaging across
        # samples when we groupby year
        if self._target_datasets is None:
            self._target_datasets = []
            for i_sample in range(target_ds.sizes["sample"]):
                self._target_datasets.append(
                    target_ds.isel(sample=i_sample)
                    .groupby(target_ds["valid_time"].isel(sample=i_sample).dt.year)
                    .sum()
                )
        else:
            for i_sample in range(target_ds.sizes["sample"]):
                self._target_datasets[i_sample] = _add_dataarray(
                    self._target_datasets[i_sample],
                    target_ds.isel(sample=i_sample)
                    .groupby(target_ds["valid_time"].isel(sample=i_sample).dt.year)
                    .sum(),
                )

        if self._gen_datasets is None:
            self._gen_datasets = []
            for i_sample in range(gen_ds.sizes["sample"]):
                self._gen_datasets.append(
                    gen_ds.isel(sample=i_sample)
                    .groupby(gen_ds["valid_time"].isel(sample=i_sample).dt.year)
                    .sum()
                )
        else:
            for i_sample in range(gen_ds.sizes["sample"]):
                self._gen_datasets[i_sample] = _add_dataarray(
                    self._gen_datasets[i_sample],
                    gen_ds.isel(sample=i_sample)
                    .groupby(gen_ds["valid_time"].isel(sample=i_sample).dt.year)
                    .sum(),
                )

    def _get_gathered_means(self) -> Optional[Tuple[xr.Dataset, xr.Dataset]]:
        """
        Gather the mean target and generated data across all processes.

        Returns:
            A tuple of the target and generated datasets, or None if this is not the
            root rank.
        """
        if self._target_datasets is None or self._gen_datasets is None:
            raise ValueError("No data has been recorded yet.")
        dist = Distributed.get_instance()
        target = xr.concat(self._target_datasets, dim="sample")
        gen = xr.concat(self._gen_datasets, dim="sample")
        if dist.world_size > 1:
            target = _gather_sample_datasets(dist, target)
            gen = _gather_sample_datasets(dist, gen)
        if target is None or gen is None:
            return None  # we are not root rank
        # filter out data with insufficient samples
        min_samples = _get_min_samples(self.timestep)
        target = target.where(target["counts"] > min_samples, drop=True)
        gen = gen.where(gen["counts"] > min_samples, drop=True)
        target = target / target["counts"]
        gen = gen / gen["counts"]
        return target, gen

    @torch.no_grad()
    def get_logs(self, label: str) -> Dict[str, Any]:
        gathered = self._get_gathered_means()
        if gathered is None:  # not the root rank
            return {}
        target, gen = gathered
        plots = {}
        metrics = {}

        for name in gen.data_vars.keys():
            if name == "counts":
                continue
            fig = Figure()  # create directly for cleanup when it leaves scope
            ax = fig.add_subplot(1, 1, 1)  # Add an axes to the figure
            ref = self._get_reference(name)
            if ref is not None:
                if ref.mean.sizes["year"] > 1:
                    # dataarray.plot() does not work for singleton dimensions
                    ref.mean.plot(ax=ax, x="year", label="ref_mean", color="black")
                    ref.min.plot(
                        ax=ax, x="year", label="ref_min", color="grey", linestyle="--"
                    )
                    ref.max.plot(
                        ax=ax, x="year", label="ref_max", color="grey", linestyle="--"
                    )
            if gen.sizes["year"] > 1:
                target_ensemble_mean = target[name].mean("sample")
                gen_ensemble_mean = gen[name].mean("sample")
                # compute R2 values
                if ref is not None:
                    r2_target = get_r2(target_ensemble_mean, ref.mean)
                    r2_gen = get_r2(gen_ensemble_mean, ref.mean)
                    metrics[f"r2/{name}_target"] = r2_target
                    metrics[f"r2/{name}_gen"] = r2_gen
                    target_label = f"target R2: {r2_target:.2f}"
                    gen_label = f"gen R2: {r2_gen:.2f}"
                else:
                    target_label = "target"
                    gen_label = "gen"
                if "year" in target_ensemble_mean.dims:
                    target_ensemble_mean.plot(
                        ax=ax, x="year", label=target_label, color="orange"
                    )
                if "year" in gen_ensemble_mean.dims:
                    gen_ensemble_mean.plot(
                        ax=ax, x="year", label=gen_label, color="cornflowerblue"
                    )
                if "year" in gen[name].dims:
                    for i_sample in range(gen.sizes["sample"]):
                        gen[name].isel(sample=i_sample).plot(
                            ax=ax,
                            x="year",
                            color="cornflowerblue",
                            alpha=0.4,
                            linestyle="-.",
                            marker="x",
                        )
            ax.set_title(f"{name}")
            ax.legend()
            fig.tight_layout()
            plots[name] = fig
        if len(label) > 0:
            label = label + "/"
        logs = {}
        logs.update({f"{label}{name}": plots[name] for name in plots.keys()})
        logs.update({f"{label}{name}": metrics[name] for name in metrics.keys()})
        return logs

    def get_dataset(self) -> xr.Dataset:
        gathered = self._get_gathered_means()
        if gathered is None:
            return xr.Dataset()
        target, gen = gathered
        return xr.concat(
            [
                target.expand_dims({"source": ["target"]}),
                gen.expand_dims({"source": ["prediction"]}),
            ],
            dim="source",
        )


@dataclasses.dataclass
class VariableReferenceData:
    mean: xr.DataArray
    min: xr.DataArray
    max: xr.DataArray


def process_monthly_reference(
    monthly_reference_data: xr.Dataset,
    area_weighted_mean: Callable[[torch.Tensor], torch.Tensor],
    name: str,
) -> VariableReferenceData:
    ref_global_mean = xr.DataArray(
        area_weighted_mean(torch.as_tensor(monthly_reference_data[name].values)),
        dims=monthly_reference_data[name].dims[:-2],
        coords={"time": monthly_reference_data[name].coords["time"]},
    )
    valid_time_0 = monthly_reference_data.valid_time.isel(sample=0)
    for i in range(1, monthly_reference_data.sizes["sample"]):
        valid_time_i = monthly_reference_data.valid_time.isel(sample=i)
        if not valid_time_0.equals(valid_time_i):
            raise ValueError("All time axes must be the same")
    # need to select just one time axis so we don't lose sample dim
    ref_annual_coarsened = (ref_global_mean * monthly_reference_data["counts"]).groupby(
        valid_time_0.dt.year
    ).sum() / monthly_reference_data["counts"].groupby(valid_time_0.dt.year).sum()
    return VariableReferenceData(
        mean=ref_annual_coarsened.mean("sample"),
        min=ref_annual_coarsened.min("sample"),
        max=ref_annual_coarsened.max("sample"),
    )


def _add_dataarray(da1: xr.DataArray, da2: xr.DataArray):
    """
    Perform dataarray addition, assuming any missing year indices
    have zero values.
    """
    union_index = np.union1d(da1.year.values, da2.year.values)
    da1 = da1.reindex(year=union_index, fill_value=0)
    da2 = da2.reindex(year=union_index, fill_value=0)
    return da1 + da2


def get_r2(da: xr.DataArray, reference: xr.DataArray) -> float:
    """Compute the R2 value of the target compared to the reference."""
    ref_data = reference.sel(year=da.year)
    SS_ref = np.sum((ref_data.values - np.mean(ref_data.values)) ** 2)
    SS_pred = np.sum((da - ref_data).values ** 2)
    return float(1 - SS_pred / SS_ref)


def _gather_sample_datasets(
    dist: Distributed, dataset: xr.Dataset
) -> Optional[xr.Dataset]:
    """
    Gather the dataset across all processes, concatenating on the sample dimension.

    Assumes all dataset variables have the same dimensions and shape, and that the
    first dimension is "sample".
    """
    # collect all data into one torch.Tensor for gathering, sort for determinism
    names = sorted(list(dataset.data_vars))
    tensor = torch.cat(
        [torch.asarray(np.expand_dims(dataset[name].values, axis=0)) for name in names],
        dim=0,
    ).to(get_device())
    years = torch.asarray(dataset.year.values).to(get_device())
    gathered_tensors = dist.gather_irregular(tensor)
    gathered_years = dist.gather_irregular(years)
    if gathered_tensors is None or gathered_years is None:
        return None
    datasets = []
    for tensor, years in zip(gathered_tensors, gathered_years):
        single_rank_dataset = xr.Dataset(
            {
                name: (["sample", "year"], tensor[i].cpu())
                for i, name in enumerate(names)
            },
            coords={"year": years.cpu()},
        )
        datasets.append(single_rank_dataset)
    # concat ranks along sample dim
    dataset_out = xr.concat(datasets, dim="sample")
    return dataset_out


@torch.no_grad()
def to_dataset(data: TensorMapping, time: xr.DataArray) -> xr.Dataset:
    """Convert a dictionary of data to an xarray dataset."""
    assert time.dims == ("sample", "time")  # must be consistent with this module
    data_vars = {}
    for name, tensor in data.items():
        data_vars[name] = (["sample", "time"], tensor)
    data_vars["counts"] = (["sample", "time"], np.ones(shape=time.shape))
    return xr.Dataset(data_vars, coords={"valid_time": time})


def _get_min_samples(timestep: datetime.timedelta) -> int:
    steps_per_day = datetime.timedelta(days=1) // timestep
    return 362 * steps_per_day



import pytest
import torch

from fme.core import metrics
from fme.core.device import get_device
from fme.core.gridded_ops import LatLonOperations
from fme.core.loss import (
    AreaWeightedMSELoss,
    GlobalMeanLoss,
    LossConfig,
    MappingLoss,
    VariableWeightingLoss,
    WeightedMappingLossConfig,
    _construct_weight_tensor,
)
from fme.core.packer import Packer


@pytest.mark.parametrize("global_mean_type", [None, "LpLoss"])
def test_loss_builds_and_runs(global_mean_type):
    config = LossConfig(global_mean_type=global_mean_type)
    area = torch.randn(10, 10, device=get_device())
    loss = config.build(LatLonOperations(area).area_weighted_mean, reduction="mean")
    x = torch.randn(10, 10, 10, 10, 10, device=get_device())
    y = torch.randn(10, 10, 10, 10, 10, device=get_device())
    result = loss(x, y)
    assert result.shape == ()
    assert isinstance(result, torch.Tensor)


def test_loss_of_zeros_is_variance():
    torch.manual_seed(0)
    config = LossConfig(global_mean_type=None)
    area = torch.randn(10, 10, device=get_device())
    loss = config.build(LatLonOperations(area).area_weighted_mean, reduction="mean")
    x = torch.zeros(10, 10, 10, 10, 10, device=get_device())
    y = torch.randn(10, 10, 10, 10, 10, device=get_device())
    result = loss(x, y)
    assert result.shape == ()
    assert isinstance(result, torch.Tensor)
    tol = {"rtol": 1e-4, "atol": 1e-4} if str(get_device()).startswith("cuda") else {}
    torch.testing.assert_close(result, y.var(), **tol)


@pytest.mark.parametrize("global_mean_weight", [0.0, 1.0, 5.0])
def test_loss_of_zeros_is_one_plus_global_mean_weight(global_mean_weight: float):
    torch.manual_seed(0)
    config = LossConfig(
        global_mean_type="LpLoss", global_mean_weight=global_mean_weight
    )
    area = torch.randn(10, 10, device=get_device())
    loss = config.build(LatLonOperations(area).area_weighted_mean, reduction="mean")
    x = torch.zeros(10, 10, 10, 10, 10, device=get_device())
    y = torch.randn(10, 10, 10, 10, 10, device=get_device())
    result = loss(x, y)
    assert result.shape == ()
    assert isinstance(result, torch.Tensor)
    expected = torch.tensor(1.0 + global_mean_weight)
    tol = (
        {"atol": 0.015, "rtol": 0.01}
        if str(get_device()).startswith("cuda")
        else {"atol": 0.01, "rtol": 0.0}
    )
    torch.testing.assert_close(result.cpu(), expected, **tol)


def test_global_mean_loss():
    torch.manual_seed(0)
    area = torch.randn(10, 10, device=get_device())
    loss = GlobalMeanLoss(
        LatLonOperations(area).area_weighted_mean, loss=torch.nn.MSELoss()
    )
    x = torch.zeros(10, 10, 10, 10, 10, device=get_device())
    y = torch.randn(10, 10, 10, 10, 10, device=get_device())
    result = loss(x, y)
    assert result.shape == ()
    assert isinstance(result, torch.Tensor)

    def global_weighted_mean(tensor, area):
        return (tensor * area[None, None, None, :, :]).sum(dim=(-1, -2)) / area.sum()

    mse = torch.nn.MSELoss()
    expected = mse(global_weighted_mean(x, area), global_weighted_mean(y, area))
    torch.testing.assert_close(result, expected)


def test_area_weighted_mse():
    torch.manual_seed(0)
    x = torch.rand(10, 10).to(get_device())
    target = torch.rand(10, 10).to(get_device())
    area = torch.rand(10, 10).to(get_device())
    area_weighted_mse = AreaWeightedMSELoss(LatLonOperations(area).area_weighted_mean)
    result = area_weighted_mse(x, target)
    expected = metrics.weighted_mean(
        torch.nn.MSELoss(reduction="none")(x, target), weights=area, dim=(-2, -1)
    ).mean()
    torch.testing.assert_close(result, expected)


def test__construct_weight_tensor():
    out_names = ["a", "b", "c"]
    weights = {"a": 0.5, "c": 3.0}
    n_samples = 10
    nlat, nlon = 8, 4
    gen_data = torch.rand(n_samples, len(out_names), nlat, nlon, device=get_device())
    weight_tensor = _construct_weight_tensor(
        weights, out_names, n_dim=4, channel_dim=-3
    )
    weighted_gen_data = gen_data * weight_tensor
    assert weight_tensor.shape == (1, len(out_names), 1, 1)
    assert weighted_gen_data.shape == gen_data.shape
    for i, name in enumerate(out_names):
        if name in weights:
            assert torch.allclose(
                weighted_gen_data[:, i], weights[name] * gen_data[:, i]
            )
        else:
            assert torch.allclose(weighted_gen_data[:, i], gen_data[:, i])


def test__construct_weight_tensor_missing_key_error():
    out_names = ["a", "b", "c"]
    weights = {"a": 0.5, "c": 3.0, "d": 1.5}
    with pytest.raises(KeyError):
        _construct_weight_tensor(weights, out_names, n_dim=4, channel_dim=-3)(
            out_names, n_dim=4, channel_dim=-3
        )


def test_MappingLoss():
    loss = torch.nn.MSELoss()
    n_channels = 5
    packer = Packer([f"var_{i}" for i in range(n_channels)])
    mapping_loss = MappingLoss(loss, packer)
    x = torch.randn(
        15,
        n_channels,
        10,
        10,
    ).to(get_device(), dtype=torch.float)
    y = torch.randn(
        15,
        n_channels,
        10,
        10,
    ).to(get_device(), dtype=torch.float)
    x_mapping = {name: x[:, i, :, :] for i, name in enumerate(packer.names)}
    y_mapping = {name: y[:, i, :, :] for i, name in enumerate(packer.names)}
    assert torch.allclose(mapping_loss(x_mapping, y_mapping), loss(x, y))


def test_VariableWeightingLoss():
    weights = torch.tensor(
        [
            4.0,
            1.0,
        ]
    ).to(get_device())
    mse_loss = torch.nn.MSELoss()
    weighted_loss = VariableWeightingLoss(weights=weights, loss=mse_loss)

    x = torch.tensor([[1.0, 2.0]]).to(get_device())
    y = torch.tensor([[2, 2.5]]).to(get_device())

    weighted_result = weighted_loss(x, y)
    assert weighted_result == ((16 + 0.25) / 2.0)


def test_WeightedMappingLossConfig_no_weights():
    loss_config = LossConfig()
    n_channels = 5
    out_names = [f"var_{i}" for i in range(n_channels)]
    channel_dim = -3
    area = torch.tensor([])  # area not used by this config
    area_weighted_mean = LatLonOperations(area).area_weighted_mean
    mapping_loss_config = WeightedMappingLossConfig()
    loss = loss_config.build(area_weighted_mean, reduction="mean")
    mapping_loss = mapping_loss_config.build(area_weighted_mean, out_names, channel_dim)
    packer = Packer(out_names)

    x_mapping = {name: torch.randn(4, 5, 5).to(get_device()) for name in out_names}
    y_mapping = {name: torch.randn(4, 5, 5).to(get_device()) for name in out_names}
    x = packer.pack(x_mapping, axis=channel_dim)
    y = packer.pack(y_mapping, axis=channel_dim)

    assert loss(x, y) == mapping_loss(x_mapping, y_mapping)


def test_WeightedMappingLossConfig_weights():
    out_names = ["var_0", "var_1"]
    channel_dim = -3
    area = torch.tensor([])  # area not used by this config
    mapping_loss_config = WeightedMappingLossConfig(
        type="MSE", weights={"var_0": 4.0, "var_1": 1.0}
    )
    mapping_loss = mapping_loss_config.build(
        LatLonOperations(area).area_weighted_mean, out_names, channel_dim
    )

    x0 = torch.ones(4, 5, 5).to(get_device())
    x1 = 2.0 * x0
    y0 = 2.0 * x0
    y1 = 2.5 * x0

    x_mapping = {"var_0": x0, "var_1": x1}
    y_mapping = {"var_0": y0, "var_1": y1}

    assert mapping_loss(x_mapping, y_mapping) == ((16 + 0.25) / 2.0)



import abc
import datetime
from typing import Any, Mapping, Protocol

from fme.core.coordinates import HybridSigmaPressureCoordinate
from fme.core.gridded_ops import GriddedOperations
from fme.core.typing_ import TensorMapping


class CorrectorConfigProtocol(Protocol):
    def build(
        self,
        gridded_operations: GriddedOperations,
        vertical_coordinate: HybridSigmaPressureCoordinate,
        timestep: datetime.timedelta,
    ) -> "CorrectorABC": ...

    @classmethod
    def from_state(cls, state: Mapping[str, Any]) -> "CorrectorConfigProtocol":
        """
        Create a ModuleSelector from a dictionary containing all the information
        needed to build a ModuleConfig.
        """
        ...


class CorrectorABC(abc.ABC):
    @abc.abstractmethod
    def __call__(
        self,
        input_data: TensorMapping,
        gen_data: TensorMapping,
        forcing_data: TensorMapping,
    ) -> TensorMapping: ...



from typing import Dict, Mapping, Optional

import torch

from fme.core.dataset.data_typing import VariableMetadata
from fme.core.typing_ import TensorMapping
from fme.core.wandb import Image

from ..plotting import plot_paneled_data


class SnapshotAggregator:
    """
    An aggregator that records the first sample of the last batch of data.
    """

    _captions = {
        "full-field": (
            "{name} one step full field for last sample; "
            "(top) generated and (bottom) target [{units}]"
        ),
        "residual": (
            "{name} one step residual (prediction - previous time) for last sample; "
            "(top) generated and (bottom) target [{units}]"
        ),
        "error": (
            "{name} one step full field error (generated - target) "
            "for last sample [{units}]"
        ),
    }

    def __init__(self, metadata: Optional[Mapping[str, VariableMetadata]] = None):
        """
        Args:
            metadata: Mapping of variable names their metadata that will
                used in generating logged image captions.
        """
        if metadata is None:
            self._metadata: Mapping[str, VariableMetadata] = {}
        else:
            self._metadata = metadata

    @torch.no_grad()
    def record_batch(
        self,
        loss: float,
        target_data: TensorMapping,
        gen_data: TensorMapping,
        target_data_norm: TensorMapping,
        gen_data_norm: TensorMapping,
    ):
        self._loss = loss
        self._target_data = target_data
        self._gen_data = gen_data
        self._target_data_norm = target_data_norm
        self._gen_data_norm = gen_data_norm

    @torch.no_grad()
    def get_logs(self, label: str) -> Dict[str, Image]:
        """
        Returns logs as can be reported to WandB.

        Args:
            label: Label to prepend to all log keys.
        """
        time_dim = 1
        input_time = 0
        target_time = 1
        image_logs = {}
        for name in self._gen_data.keys():
            # use first sample in batch
            gen = self._gen_data[name].select(dim=time_dim, index=target_time)[0].cpu()
            target = (
                self._target_data[name].select(dim=time_dim, index=target_time)[0].cpu()
            )
            input = (
                self._target_data[name].select(dim=time_dim, index=input_time)[0].cpu()
            )
            images = {}
            images["error"] = [[(gen - target).numpy()]]
            images["full-field"] = [[gen.numpy()], [target.numpy()]]
            images["residual"] = [[(gen - input).numpy()], [(target - input).numpy()]]
            for key, data in images.items():
                if key == "error" or key == "residual":
                    diverging = True
                else:
                    diverging = False
                caption = self._get_caption(key, name)
                wandb_image = plot_paneled_data(data, diverging, caption=caption)
                image_logs[f"image-{key}/{name}"] = wandb_image
        image_logs = {f"{label}/{key}": image_logs[key] for key in image_logs}
        return image_logs

    def _get_caption(self, key: str, name: str) -> str:
        if name in self._metadata:
            caption_name = self._metadata[name].long_name
            units = self._metadata[name].units
        else:
            caption_name, units = name, "unknown_units"
        caption = self._captions[key].format(name=caption_name, units=units)
        return caption



import numpy as np
import pytest
import torch
import torch_harmonics

import fme
from fme.core.coordinates import HybridSigmaPressureCoordinate
from fme.core.metrics import (
    net_surface_energy_flux,
    quantile,
    spherical_power_spectrum,
    surface_pressure_due_to_dry_air,
)


def _get_lats(num_lat: int):
    """A simple example of latitudes."""
    return torch.linspace(-89.5, 89.5, num_lat)


test_cases = (
    "variable, time, lats, n_lon",
    [
        (2, 1, _get_lats(2), 1),
        (1, 2, _get_lats(2), 1),
        (1, 2, _get_lats(2), 4),
    ],
)


@pytest.mark.parametrize(
    "num_lat, num_lon, expected",
    [
        (2, 1, torch.tensor([[0.5], [0.5]])),
        (2, 2, torch.tensor([[0.25, 0.25], [0.25, 0.25]])),
        (
            2,
            4,
            torch.tensor(
                [[0.1250, 0.1250, 0.1250, 0.1250], [0.1250, 0.1250, 0.1250, 0.1250]]
            ),
        ),
    ],
)
def test_spherical_area_weights(num_lat, num_lon, expected):
    """Tests the shapes and a couple simple cases of the spherical area weights."""
    lats = _get_lats(num_lat)
    result = fme.spherical_area_weights(lats, num_lon)
    assert torch.all(torch.isclose(result, expected))


@pytest.mark.parametrize(*test_cases)
def test_weighted_mean(variable, time, lats, n_lon):
    """Tests the weighted mean for a few simple test cases."""
    x = torch.randn(time, variable, len(lats), n_lon)
    weights = fme.spherical_area_weights(lats, n_lon)

    result = fme.weighted_mean(x, weights, dim=(0, 2, 3))
    assert result.shape == (variable,), "You should be able to specify time as dim = 1."

    result = fme.weighted_mean(
        torch.zeros(variable, time, len(lats), n_lon), weights, dim=(0, 2, 3)
    )
    assert torch.all(
        torch.isclose(result, torch.tensor([0.0]))
    ), "Weighted mean of zeros should be zero."

    result = fme.weighted_mean(torch.ones(variable, time, len(lats), n_lon), weights)
    assert torch.all(
        torch.isclose(result, torch.Tensor([1.0]))
    ), "The weighted mean of a constant should be that constant."


def test_weighted_std_constant_weights():
    """Tests the weighted std with constant weights."""
    torch.manual_seed(0)
    n_lat = 4
    n_lon = 5
    x = torch.randn(2, 3, n_lat, n_lon)
    weights = torch.ones(n_lat, n_lon)
    result = fme.weighted_std(x, weights, dim=(0, 2, 3))
    # must use numpy std here as we want the standard deviation of
    # the sample, not the population
    unweighted_std = np.std(x.cpu().numpy(), axis=(0, 2, 3))
    assert torch.all(
        torch.isclose(result, torch.as_tensor(unweighted_std))
    ), "Weighted std with constant weights should be same as unweighted std."


def test_weighted_std_none_weights():
    """Tests the weighted std without passing weights."""
    torch.manual_seed(0)
    n_lat = 4
    n_lon = 5
    x = torch.randn(2, 3, n_lat, n_lon)
    weights = None
    result = fme.weighted_std(x, weights, dim=(0, 2, 3))
    # must use numpy std here as we want the standard deviation of
    # the sample, not the population
    unweighted_std = np.std(x.cpu().numpy(), axis=(0, 2, 3))
    assert torch.all(
        torch.isclose(result, torch.as_tensor(unweighted_std))
    ), "Weighted std with constant weights should be same as unweighted std."


@pytest.mark.parametrize(*test_cases)
def test_weighted_mean_bias(variable, time, lats, n_lon):
    """Tests the weighted mean bias for a few simple test cases."""
    x = torch.randn(time, variable, len(lats), n_lon)
    y = torch.randn(time, variable, len(lats), n_lon)
    weights = fme.spherical_area_weights(lats, n_lon)

    result = fme.weighted_mean_bias(x, x.clone(), weights, dim=(0, 2, 3))
    assert torch.all(
        torch.isclose(result, torch.tensor(0.0))
    ), "Weighted global mean bias between identical tensors should be zero."
    assert result.shape == (variable,), "You should be able to specify time as dim = 1."

    x = torch.zeros(time, variable, len(lats), n_lon)
    y = torch.ones(time, variable, len(lats), n_lon)

    result = fme.weighted_mean_bias(x, y, weights)
    assert torch.all(
        torch.isclose(result, torch.Tensor([1.0]))
    ), "The weighted mean of a constant should be that constant."

    result = fme.weighted_mean_bias(x, y)
    assert result.shape == tuple(), "Should also work if you do not specify weights."

    x = torch.randn(variable, time, n_lon, len(lats))
    y = torch.randn(variable, time, n_lon, len(lats))
    result = fme.weighted_mean_bias(x, x.clone(), weights.t(), dim=(1, 2, 3))
    assert torch.all(
        torch.isclose(result, torch.tensor(0.0))
    ), "Weighted global mean bias between identical tensors should be zero."
    assert result.shape == (
        variable,
    ), "Swapping dims shouldn't change the final shape."


@pytest.mark.parametrize(*test_cases)
def test_root_mean_squared_error(variable, time, lats, n_lon):
    """Tests the mean squared error for a few simple test cases."""
    x = torch.randn(variable, time, len(lats), n_lon)
    random_weights = torch.rand(len(lats), n_lon)

    result = fme.root_mean_squared_error(x, x.clone(), dim=(0, 2, 3))
    assert torch.all(
        torch.isclose(result, torch.tensor(0.0))
    ), "Root mean squared error between identical tensors should be zero."

    result = fme.root_mean_squared_error(
        torch.zeros(variable, time, len(lats), n_lon),
        torch.ones(variable, time, len(lats), n_lon),
    )
    assert torch.all(
        torch.isclose(result, torch.tensor(1.0))
    ), "Root mean squared error between zero and one should be one."

    result = fme.root_mean_squared_error(
        torch.zeros(variable, time, len(lats), n_lon),
        torch.ones(variable, time, len(lats), n_lon),
        weights=random_weights,
    )
    assert torch.isclose(
        result, torch.tensor([1.0]).sqrt()
    ), "Root mean squared error between zero and one should be one."


@pytest.mark.parametrize(*test_cases)
def test_rmse_of_time_mean(variable, time, lats, n_lon):
    x = torch.randn(variable, time, len(lats), n_lon)
    random_weights = torch.rand(len(lats), n_lon)

    result = fme.rmse_of_time_mean(x, x.clone(), time_dim=1)
    torch.testing.assert_close(
        result,
        torch.zeros((variable,)),
        msg="RMSE of time mean between identical tensors should be zero.",
    )

    result = fme.rmse_of_time_mean(
        torch.zeros(variable, time, len(lats), n_lon),
        torch.ones(variable, time, len(lats), n_lon),
        weights=random_weights,
        time_dim=1,
    )
    torch.testing.assert_close(
        result,
        torch.ones((variable,)),
        msg="RMSE of time mean between zero and 1 should be 1.",
    )


@pytest.mark.parametrize(*test_cases)
def test_time_and_global_mean(variable, time, lats, n_lon):
    x = torch.randn(variable, time, len(lats), n_lon)
    random_weights = torch.rand(len(lats), n_lon)

    result = fme.time_and_global_mean_bias(x, x.clone(), time_dim=1)
    torch.testing.assert_close(
        result,
        torch.zeros((variable,)),
        msg="Time and global mean bias between identical tensors should be zero.",
    )

    result = fme.time_and_global_mean_bias(
        torch.zeros(variable, time, len(lats), n_lon),
        torch.ones(variable, time, len(lats), n_lon),
        weights=random_weights,
        time_dim=1,
    )
    torch.testing.assert_close(
        result,
        torch.ones((variable,)),
        msg="Global mean bias between zero and 1 should be 1.",
    )


def test_gradient_magnitude():
    constant = torch.ones((5, 2, 4, 4))
    constant_grad_magnitude = fme.gradient_magnitude(constant, dim=(-2, -1))
    torch.testing.assert_close(constant_grad_magnitude, torch.zeros_like(constant))

    monotonic = torch.tile(torch.arange(4.0), (5, 2, 4, 1))
    monotonic_grad_magnitude = fme.gradient_magnitude(monotonic, dim=(-2, -1))
    torch.testing.assert_close(monotonic_grad_magnitude, torch.ones_like(monotonic))


def test_weighted_mean_gradient_magnitude():
    constant = torch.ones((5, 2, 4, 4))
    constant_grad_magnitude = fme.weighted_mean_gradient_magnitude(
        constant, dim=(-2, -1)
    )
    torch.testing.assert_close(constant_grad_magnitude, torch.zeros((5, 2)))

    monotonic = torch.tile(torch.arange(4.0), (5, 2, 4, 1))
    monotonic_grad_magnitude = fme.weighted_mean_gradient_magnitude(
        monotonic, dim=(-2, -1)
    )
    torch.testing.assert_close(monotonic_grad_magnitude, torch.ones((5, 2)))


def test_gradient_magnitude_percent_diff():
    constant = torch.ones((5, 2, 4, 4))
    self_percent_diff = fme.gradient_magnitude_percent_diff(
        constant, constant, dim=(-2, -1)
    )
    assert torch.all(torch.isnan(self_percent_diff))

    monotonic = torch.tile(torch.arange(4.0), (5, 2, 4, 1))
    percent_diff = fme.gradient_magnitude_percent_diff(
        monotonic, constant, dim=(-2, -1)
    )
    torch.testing.assert_close(percent_diff, -100 * torch.ones((5, 2)))


def test_dry_air_shapes():
    nlat, nlon, nz = 4, 8, 3
    water = torch.rand(nlat, nlon, nz)
    pressure = torch.rand(nlat, nlon)
    ak, bk = torch.arange(nz + 1), torch.arange(nz + 1)
    coords = HybridSigmaPressureCoordinate(ak, bk)
    dry_air = surface_pressure_due_to_dry_air(water, pressure, coords)
    assert dry_air.shape == (nlat, nlon)


def single_level_ak_bk():
    ak = torch.zeros(size=[2])
    bk = torch.asarray([0.0, 1.0])
    return ak, bk


def test_single_level_dry_air_no_water():
    torch.manual_seed(0)
    nlat, nlon, nz = 4, 8, 1
    water = torch.zeros(nlat, nlon, nz)
    pressure = torch.rand(nlat, nlon)
    ak, bk = single_level_ak_bk()
    coords = HybridSigmaPressureCoordinate(ak, bk)
    dry_air = surface_pressure_due_to_dry_air(water, pressure, coords)
    np.testing.assert_allclose(dry_air.cpu().numpy(), pressure.cpu().numpy())


def test_single_level_dry_air_all_water():
    torch.manual_seed(0)
    nlat, nlon, nz = 4, 8, 1
    water = torch.ones(nlat, nlon, nz)
    pressure = torch.rand(nlat, nlon)
    ak, bk = single_level_ak_bk()
    coords = HybridSigmaPressureCoordinate(ak, bk)
    dry_air = surface_pressure_due_to_dry_air(water, pressure, coords)
    np.testing.assert_almost_equal(dry_air.cpu().numpy(), 0.0, decimal=6)


def test_single_level_dry_air_some_water():
    torch.manual_seed(0)
    nlat, nlon, nz = 4, 8, 1
    water = torch.rand(nlat, nlon, nz)
    pressure = torch.rand(nlat, nlon)
    ak, bk = single_level_ak_bk()
    target_dry_air = pressure * (1.0 - water[:, :, 0])
    coords = HybridSigmaPressureCoordinate(ak, bk)
    dry_air = surface_pressure_due_to_dry_air(water, pressure, coords)
    np.testing.assert_allclose(
        dry_air.cpu().numpy(), target_dry_air.cpu().numpy(), rtol=1e-5
    )


def test_net_surface_energy_flux():
    sfc_down_lw_radiative_flux = torch.tensor([100.0])
    sfc_up_lw_radiative_flux = torch.tensor([50.0])
    sfc_down_sw_radiative_flux = torch.tensor([25.0])
    sfc_up_sw_radiative_flux = torch.tensor([10.0])
    latent_heat_flux = torch.tensor([5.0])
    sensible_heat_flux = torch.tensor([2.5])
    expected_net_surface_energy_flux = torch.tensor([57.5])
    result = net_surface_energy_flux(
        sfc_down_lw_radiative_flux,
        sfc_up_lw_radiative_flux,
        sfc_down_sw_radiative_flux,
        sfc_up_sw_radiative_flux,
        latent_heat_flux,
        sensible_heat_flux,
    )
    torch.testing.assert_close(result, expected_net_surface_energy_flux)


@pytest.mark.parametrize(
    "bins, hist, pct, expected",
    [
        pytest.param(
            np.array([0, 1]), np.array([10]), 0.8, 0.8, id="single_bin_interpolation"
        ),
        pytest.param(
            np.array([0, 1, 2]),
            np.array([10, 10]),
            0.8,
            1.6,
            id="two_bin_interpolation",
        ),
        pytest.param(
            np.array([0, 1, 2]), np.array([10, 10]), 1.0, 2.0, id="two_bin_end_value"
        ),
    ],
)
def test_quantile(bins, hist, pct, expected):
    result = quantile(bins, hist, pct)
    np.testing.assert_allclose(result, expected)


def test_spherical_power_spectrum():
    nlat, nlon = 4, 8
    data = torch.rand(2, nlat, nlon)
    sht = torch_harmonics.RealSHT(nlat, nlon)
    spectrum = spherical_power_spectrum(data, sht)
    assert spectrum.shape == (2, nlat - 1)



import copy
import dataclasses
import pathlib
import subprocess
import tempfile
import textwrap
import unittest.mock

import numpy as np
import pytest
import torch
import xarray as xr
import yaml

from fme.ace.inference.evaluator import main as inference_evaluator_main
from fme.ace.registry.test_hpx import (
    conv_next_block_config,
    decoder_config,
    down_sampling_block_config,
    encoder_config,
    output_layer_config,
    recurrent_block_config,
    up_sampling_block_config,
)
from fme.ace.testing import (
    DimSizes,
    MonthlyReferenceData,
    save_nd_netcdf,
    save_scalar_netcdf,
)
from fme.ace.train.train import main as train_main
from fme.core.coordinates import (
    HEALPixCoordinates,
    HorizontalCoordinates,
    LatLonCoordinates,
)
from fme.core.generics.trainer import (
    _restore_checkpoint,
    count_parameters,
    epoch_checkpoint_enabled,
)
from fme.core.testing.wandb import mock_wandb
from fme.core.typing_ import Slice

REPOSITORY_PATH = pathlib.PurePath(__file__).parent.parent.parent.parent
JOB_SUBMISSION_SCRIPT_PATH = (
    REPOSITORY_PATH / "fme" / "fme" / "ace" / "run-train-and-inference.sh"
)


def _get_test_yaml_files(
    *,
    train_data_path,
    valid_data_path,
    monthly_data_filename,
    results_dir,
    global_means_path,
    global_stds_path,
    in_variable_names,
    out_variable_names,
    mask_name,
    n_forward_steps=2,
    nettype="SphericalFourierNeuralOperatorNet",
    stepper_checkpoint_file=None,
    log_to_wandb=False,
    max_epochs=1,
    segment_epochs=1,
    inference_forward_steps=2,
    use_healpix=False,
):
    input_time_size = 1
    output_time_size = 1
    if use_healpix:
        in_channels = len(in_variable_names)
        out_channels = len(out_variable_names)
        prognostic_variables = min(
            out_channels, in_channels
        )  # how many variables in/out share.
        # in practice, we will need to compare variable names, since there
        # are some input-only and some output-only channels.
        # TODO: https://github.com/ai2cm/full-model/issues/1046
        n_constants = 0
        decoder_input_channels = 0  # was 1, to indicate insolation - now 0
        input_time_size = 1  # TODO: change to 2 (issue #1177)
        output_time_size = 1  # TODO: change to 4 (issue #1177)
        spatial_dimensions_str = "healpix"

        conv_next_block = conv_next_block_config(in_channels=in_channels)
        down_sampling_block = down_sampling_block_config()
        recurrent_block = recurrent_block_config()
        encoder = encoder_config(
            conv_next_block, down_sampling_block, n_channels=[16, 8, 4]
        )
        up_sampling_block = up_sampling_block_config()
        output_layer = output_layer_config()
        decoder = decoder_config(
            conv_next_block,
            up_sampling_block,
            output_layer,
            recurrent_block,
            n_channels=[4, 8, 16],
        )

        # Need to manually indent these YAML strings
        encoder_yaml = yaml.dump(
            dataclasses.asdict(encoder), default_flow_style=False, indent=2
        )
        decoder_yaml = yaml.dump(
            dataclasses.asdict(decoder), default_flow_style=False, indent=2
        )
        encoder_yaml_indented = textwrap.indent(encoder_yaml, "        ")
        decoder_yaml_indented = textwrap.indent(decoder_yaml, "        ")
        config_str = f"""
      encoder:
{encoder_yaml_indented}
      decoder:
{decoder_yaml_indented}
      prognostic_variables: {prognostic_variables}
      n_constants: {n_constants}
      decoder_input_channels: {decoder_input_channels}
      input_time_size: {input_time_size}
      output_time_size: {output_time_size}
        """
    else:
        config_str = """
      num_layers: 2
      embed_dim: 12"""
        spatial_dimensions_str = "latlon"

    new_stepper_config = f"""
  in_names: {in_variable_names}
  out_names: {out_variable_names}
  normalization:
    global_means_path: '{global_means_path}'
    global_stds_path: '{global_stds_path}'
  residual_normalization:
    global_means_path: '{global_means_path}'
    global_stds_path: '{global_stds_path}'
  loss:
    type: "MSE"
  builder:
    type: {nettype}
    config: {config_str}
  ocean:
    surface_temperature_name: {in_variable_names[0]}
    ocean_fraction_name: {mask_name}
"""
    existing_stepper_config = f"""
  checkpoint_file: {stepper_checkpoint_file}
"""

    if stepper_checkpoint_file:
        stepper_config = existing_stepper_config
    else:
        stepper_config = new_stepper_config

    train_string = f"""
train_loader:
  dataset:
    - data_path: '{train_data_path}'
      spatial_dimensions: {spatial_dimensions_str}
  batch_size: 2
  num_data_workers: 0
validation_loader:
  dataset:
    - data_path: '{valid_data_path}'
      spatial_dimensions: {spatial_dimensions_str}
  batch_size: 2
  num_data_workers: 0
optimization:
  optimizer_type: "Adam"
  lr: 0.001
  scheduler:
      type: CosineAnnealingLR
      kwargs:
        T_max: 1
stepper:
{stepper_config}
inference:
  aggregator:
    monthly_reference_data: {monthly_data_filename}
  loader:
    dataset:
      data_path: '{valid_data_path}'
      spatial_dimensions: {spatial_dimensions_str}
    start_indices:
      first: 0
      n_initial_conditions: 2
      interval: 1
  n_forward_steps: {inference_forward_steps}
  forward_steps_in_memory: 2
n_forward_steps: {n_forward_steps}
max_epochs: {max_epochs}
segment_epochs: {segment_epochs}
save_checkpoint: true
logging:
  log_to_screen: true
  log_to_wandb: {str(log_to_wandb).lower()}
  log_to_file: false
  project: fme
  entity: ai2cm
experiment_dir: {results_dir}
    """  # noqa: E501
    inference_string = f"""
experiment_dir: {results_dir}
n_forward_steps: 6
forward_steps_in_memory: 2
checkpoint_path: {results_dir}/training_checkpoints/best_ckpt.tar
data_writer:
  save_prediction_files: true
aggregator:
  log_video: true
logging:
  log_to_screen: true
  log_to_wandb: {str(log_to_wandb).lower()}
  log_to_file: false
  project: fme
  entity: ai2cm
loader:
  dataset:
    data_path: '{valid_data_path}'
    spatial_dimensions: {spatial_dimensions_str}
  start_indices:
    first: 0
    n_initial_conditions: 2
    interval: 1
    """  # noqa: E501

    with tempfile.NamedTemporaryFile(mode="w", delete=False, suffix=".yaml") as f_train:
        f_train.write(train_string)

    with tempfile.NamedTemporaryFile(
        mode="w", delete=False, suffix=".yaml"
    ) as f_inference:
        f_inference.write(inference_string)

    return f_train.name, f_inference.name


def get_sizes(
    spatial_dims: HorizontalCoordinates = LatLonCoordinates(
        lon=torch.Tensor(np.arange(32)),
        lat=torch.Tensor(np.arange(16)),
        loaded_lat_name="grid_yt",
        loaded_lon_name="grid_xt",
    ),
    n_time=3,
    nz_interface=3,
) -> DimSizes:
    return DimSizes(
        n_time=n_time,
        horizontal=copy.deepcopy(spatial_dims.loaded_sizes),
        nz_interface=nz_interface,
    )


def _setup(
    path,
    nettype,
    log_to_wandb=False,
    max_epochs=1,
    segment_epochs=1,
    n_time=10,
    timestep_days=5,
    inference_forward_steps=2,
    use_healpix=False,
):
    if not path.exists():
        path.mkdir()
    seed = 0
    np.random.seed(seed)
    in_variable_names = [
        "PRESsfc",
        "specific_total_water_0",
        "specific_total_water_1",
        "baz",
    ]
    out_variable_names = ["PRESsfc", "specific_total_water_0", "specific_total_water_1"]
    mask_name = "mask"
    all_variable_names = list(set(in_variable_names + out_variable_names))

    if use_healpix:
        hpx_coords = HEALPixCoordinates(
            face=torch.Tensor(np.arange(12)),
            width=torch.Tensor(np.arange(16)),
            height=torch.Tensor(np.arange(16)),
        )
        dim_sizes = get_sizes(spatial_dims=hpx_coords, n_time=n_time)
    else:
        dim_sizes = get_sizes(n_time=n_time)

    data_dir = path / "data"
    stats_dir = path / "stats"
    results_dir = path / "output"
    data_dir.mkdir()
    stats_dir.mkdir()
    results_dir.mkdir()
    save_nd_netcdf(
        data_dir / "data.nc",
        dim_sizes,
        variable_names=all_variable_names + [mask_name],
        timestep_days=1,
    )
    save_scalar_netcdf(
        stats_dir / "stats-mean.nc",
        variable_names=all_variable_names,
    )
    save_scalar_netcdf(
        stats_dir / "stats-stddev.nc",
        variable_names=all_variable_names,
    )

    monthly_dim_sizes: DimSizes
    if use_healpix:
        hpx_coords = HEALPixCoordinates(
            face=torch.Tensor(np.arange(12)),
            width=torch.Tensor(np.arange(16)),
            height=torch.Tensor(np.arange(16)),
        )
        monthly_dim_sizes = get_sizes(
            spatial_dims=hpx_coords, n_time=10 * 12, nz_interface=1
        )
    else:
        monthly_dim_sizes = get_sizes(n_time=10 * 12, nz_interface=1)
    monthly_reference_data = MonthlyReferenceData(
        path=data_dir,
        names=out_variable_names,
        dim_sizes=monthly_dim_sizes,
        n_ensemble=3,
    )

    train_config_filename, inference_config_filename = _get_test_yaml_files(
        train_data_path=data_dir,
        valid_data_path=data_dir,
        monthly_data_filename=monthly_reference_data.data_filename,
        results_dir=results_dir,
        global_means_path=stats_dir / "stats-mean.nc",
        global_stds_path=stats_dir / "stats-stddev.nc",
        in_variable_names=in_variable_names,
        out_variable_names=out_variable_names,
        mask_name=mask_name,
        nettype=nettype,
        log_to_wandb=log_to_wandb,
        max_epochs=max_epochs,
        segment_epochs=segment_epochs,
        inference_forward_steps=inference_forward_steps,
        use_healpix=use_healpix,
    )
    return train_config_filename, inference_config_filename


@pytest.mark.parametrize(
    "nettype", ["SphericalFourierNeuralOperatorNet", "HEALPixRecUNet", "SFNO-v0.1.0"]
)
def test_train_and_inference_inline(tmp_path, nettype, very_fast_only: bool):
    """Make sure that training and inference run without errors

    Args:
        tmp_path: pytext fixture for temporary workspace.
        nettype: parameter indicating model architecture to use.
        very_fast_only: parameter indicating whether to skip slow tests.
    """
    if very_fast_only:
        pytest.skip("Skipping non-fast tests")
    # need multi-year to cover annual aggregator
    train_config, inference_config = _setup(
        tmp_path,
        nettype,
        log_to_wandb=True,
        timestep_days=20,
        n_time=int(366 * 3 / 20 + 1),
        inference_forward_steps=int(366 * 3 / 20 / 2 - 1) * 2,  # must be even
        use_healpix=(nettype == "HEALPixRecUNet"),
    )
    # using pdb requires calling main functions directly
    with mock_wandb() as wandb:
        train_main(
            yaml_config=train_config,
        )
        wandb_logs = wandb.get_logs()

        for log in wandb_logs:
            # ensure inference time series is not logged
            assert "inference/mean/forecast_step" not in log

    # inference should not require stats files
    (tmp_path / "stats" / "stats-mean.nc").unlink()
    (tmp_path / "stats" / "stats-stddev.nc").unlink()

    with mock_wandb() as wandb:
        wandb.configure(log_to_wandb=True)
        inference_evaluator_main(yaml_config=inference_config)
        inference_logs = wandb.get_logs()

    prediction_output_path = tmp_path / "output" / "autoregressive_predictions.nc"
    best_checkpoint_path = (
        tmp_path / "output" / "training_checkpoints" / "best_ckpt.tar"
    )
    best_inference_checkpoint_path = (
        tmp_path / "output" / "training_checkpoints" / "best_inference_ckpt.tar"
    )
    assert best_checkpoint_path.exists()
    assert best_inference_checkpoint_path.exists()
    n_ic_timesteps = 1
    n_forward_steps = 6
    n_summary_steps = 1
    assert len(inference_logs) == n_ic_timesteps + n_forward_steps + n_summary_steps
    assert prediction_output_path.exists()
    ds_prediction = xr.open_dataset(prediction_output_path)
    assert np.sum(np.isnan(ds_prediction["PRESsfc"].values)) == 0
    assert np.sum(np.isnan(ds_prediction["specific_total_water_0"].values)) == 0
    assert np.sum(np.isnan(ds_prediction["specific_total_water_1"].values)) == 0
    ds_target = xr.open_dataset(tmp_path / "output" / "autoregressive_target.nc")
    assert np.sum(np.isnan(ds_target["baz"].values)) == 0


@pytest.mark.parametrize("nettype", ["SphericalFourierNeuralOperatorNet"])
def test_resume(tmp_path, nettype, very_fast_only: bool):
    """Make sure the training is resumed from a checkpoint when restarted."""
    if very_fast_only:
        pytest.skip("Skipping non-fast tests")

    mock = unittest.mock.MagicMock(side_effect=_restore_checkpoint)
    with unittest.mock.patch("fme.core.generics.trainer._restore_checkpoint", new=mock):
        train_config, _ = _setup(
            tmp_path, nettype, log_to_wandb=True, max_epochs=2, segment_epochs=1
        )
        with mock_wandb() as wandb:
            train_main(
                yaml_config=train_config,
            )
        assert min([val["epoch"] for val in wandb.get_logs() if "epoch" in val]) == 0
        assert max([val["epoch"] for val in wandb.get_logs() if "epoch" in val]) == 0
        assert not mock.called
        with mock_wandb() as wandb:
            train_main(
                yaml_config=train_config,
            )
        mock.assert_called()
        assert min([val["epoch"] for val in wandb.get_logs() if "epoch" in val]) == 1
        assert max([val["epoch"] for val in wandb.get_logs() if "epoch" in val]) == 1


@pytest.mark.parametrize("nettype", ["SphericalFourierNeuralOperatorNet"])
def test_resume_two_workers(tmp_path, nettype, skip_slow: bool, tmpdir: pathlib.Path):
    """Make sure the training is resumed from a checkpoint when restarted, using
    torchrun with NPROC_PER_NODE set to 2."""
    if skip_slow:
        # script is slow as everything is re-imported when it runs
        pytest.skip("Skipping slow tests")
    train_config, inference_config = _setup(tmp_path, nettype)
    subprocess_args = [
        JOB_SUBMISSION_SCRIPT_PATH,
        train_config,
        inference_config,
        "2",  # this makes the training run on two GPUs
    ]
    initial_process = subprocess.run(subprocess_args, cwd=tmpdir)
    initial_process.check_returncode()
    resume_subprocess_args = [
        "torchrun",
        "--nproc_per_node",
        "2",
        "-m",
        "fme.ace.train",
        train_config,
    ]
    resume_process = subprocess.run(resume_subprocess_args, cwd=tmpdir)
    resume_process.check_returncode()


def _create_fine_tuning_config(path_to_train_config_yaml: str, path_to_checkpoint: str):
    # TODO(gideond) rename to "overwrite" or something of that nature
    with open(path_to_train_config_yaml, "r") as config_file:
        config_data = yaml.safe_load(config_file)
        config_data["stepper"] = {"checkpoint_path": path_to_checkpoint}
        current_experiment_dir = config_data["experiment_dir"]
        new_experiment_dir = pathlib.Path(current_experiment_dir) / "fine_tuning"
        config_data["experiment_dir"] = str(new_experiment_dir)
        with tempfile.NamedTemporaryFile(
            mode="w", delete=False, suffix=".yaml"
        ) as new_config_file:
            new_config_file.write(yaml.dump(config_data))

    return new_config_file.name, new_experiment_dir


@pytest.mark.parametrize("nettype", ["SphericalFourierNeuralOperatorNet"])
def test_fine_tuning(tmp_path, nettype, very_fast_only: bool):
    """Check that fine tuning config runs without errors."""
    if very_fast_only:
        pytest.skip("Skipping non-fast tests")
    train_config, _ = _setup(tmp_path, nettype)

    train_main(yaml_config=train_config)

    results_dir = tmp_path / "output"
    ckpt = f"{results_dir}/training_checkpoints/best_ckpt.tar"

    fine_tuning_config, new_results_dir = _create_fine_tuning_config(train_config, ckpt)

    train_main(yaml_config=fine_tuning_config)
    assert (new_results_dir / "training_checkpoints" / "ckpt.tar").exists()


def _create_copy_weights_after_batch_config(
    path_to_train_config_yaml: str, path_to_checkpoint: str, experiment_dir: str
):
    # TODO(gideond) rename to "overwrite" or something of that nature
    with open(path_to_train_config_yaml, "r") as config_file:
        config_data = yaml.safe_load(config_file)
        config_data["stepper"]["parameter_init"] = {"weights_path": path_to_checkpoint}
        config_data["copy_weights_after_batch"] = {"include": ["*"], "exclude": []}
        config_data["experiment_dir"] = experiment_dir
        with tempfile.NamedTemporaryFile(
            mode="w", delete=False, suffix=".yaml"
        ) as new_config_file:
            new_config_file.write(yaml.dump(config_data))

    return new_config_file.name


@pytest.mark.parametrize("nettype", ["SphericalFourierNeuralOperatorNet"])
def test_copy_weights_after_batch(tmp_path, nettype, skip_slow: bool):
    """Check that fine tuning config using copy_weights_after_batch
    runs without errors."""
    if skip_slow:
        pytest.skip("Skipping slow tests")

    train_config, _ = _setup(tmp_path, nettype)

    train_main(
        yaml_config=train_config,
    )

    results_dir = tmp_path / "output"
    ckpt = f"{results_dir}/training_checkpoints/best_ckpt.tar"

    fine_tuning_config = _create_copy_weights_after_batch_config(
        train_config, ckpt, experiment_dir=str(tmp_path / "fine_tuning_dir")
    )
    train_main(yaml_config=fine_tuning_config)


@pytest.mark.parametrize(
    "checkpoint_save_epochs,expected_save_epochs",
    [(None, []), (Slice(start=-2), [2, 3]), (Slice(step=2), [0, 2])],
)
def test_epoch_checkpoint_enabled(checkpoint_save_epochs, expected_save_epochs):
    max_epochs = 4
    for i in range(max_epochs):
        if i in expected_save_epochs:
            assert epoch_checkpoint_enabled(i, max_epochs, checkpoint_save_epochs)
        else:
            assert not epoch_checkpoint_enabled(i, max_epochs, checkpoint_save_epochs)


@pytest.mark.parametrize(
    "module_list,expected_num_parameters",
    [
        (torch.nn.ModuleList([torch.nn.Linear(10, 5), torch.nn.Linear(5, 2)]), 67),
        (torch.nn.ModuleList([]), 0),
    ],
)
def test_count_parameters(module_list, expected_num_parameters):
    num_parameters = count_parameters(module_list)
    assert num_parameters == expected_num_parameters



import datetime
import pathlib

import cftime
import matplotlib.pyplot as plt
import pytest
import torch
import xarray as xr

import fme
from fme.ace.aggregator.inference.annual import GlobalMeanAnnualAggregator
from fme.ace.testing import DimSizes, MonthlyReferenceData
from fme.core.coordinates import DimSize
from fme.core.device import get_device
from fme.core.gridded_ops import LatLonOperations
from fme.core.testing import mock_distributed

TIMESTEP = datetime.timedelta(hours=6)


def test_annual_aggregator(tmpdir):
    n_lat = 16
    n_lon = 32
    # need to have two actual full years of data for plotting to get exercised
    n_sample = 2
    n_time = 365 * 4 * 2
    area_weights = torch.ones(n_lat, n_lon).to(fme.get_device())
    names = ["a"]
    horizontal = [DimSize("grid_yt", n_lat), DimSize("grid_xt", n_lon)]
    monthly_reference_data = MonthlyReferenceData(
        path=pathlib.Path(tmpdir),
        names=names,
        dim_sizes=DimSizes(
            n_time=48,
            horizontal=horizontal,
            nz_interface=1,
        ),
        n_ensemble=3,
    )
    monthly_ds = xr.open_dataset(monthly_reference_data.data_filename)
    agg = GlobalMeanAnnualAggregator(
        ops=LatLonOperations(area_weights),
        timestep=TIMESTEP,
        monthly_reference_data=monthly_ds,
    )
    target_data = {
        "a": torch.randn(n_sample, n_time, n_lat, n_lon, device=get_device())
    }
    gen_data = {"a": torch.randn(n_sample, n_time, n_lat, n_lon, device=get_device())}
    time = xr.DataArray(
        [
            [
                (
                    cftime.DatetimeProlepticGregorian(2000, 1, 1)
                    + i * datetime.timedelta(hours=6)
                )
                for i in range(n_time)
            ]
            for _ in range(n_sample)
        ],
        dims=["sample", "time"],
    )
    agg.record_batch(time, target_data, gen_data)
    logs = agg.get_logs(label="test")
    assert len(logs) > 0
    assert "test/a" in logs
    assert isinstance(logs["test/a"], plt.Figure)
    assert "test/r2/a_target" in logs


@pytest.mark.parametrize("use_mock_distributed", [False, True])
def test__get_gathered_means(use_mock_distributed):
    """Test the private _get_gathered_means method, rather than the public
    get_logs method, because the public method returns some data converted to images,
    and so that portion can't easily be tested.
    """
    n_lat = 16
    n_lon = 32
    n_sample = 2
    n_time = 365 * 4 * 2  # two years, approximately
    area_weights = torch.ones(n_lat, n_lon).to(fme.get_device())
    agg = GlobalMeanAnnualAggregator(
        ops=LatLonOperations(area_weights),
        timestep=TIMESTEP,
    )
    target_data = {
        "a": torch.randn(n_sample, n_time, n_lat, n_lon, device=get_device())
    }
    gen_data = {"a": torch.randn(n_sample, n_time, n_lat, n_lon, device=get_device())}
    time = xr.DataArray(
        [
            [
                (
                    cftime.DatetimeProlepticGregorian(
                        2000 + j, 1, 1
                    )  # add j so that years are not the same
                    + i * datetime.timedelta(hours=6)
                )
                for i in range(n_time)
            ]
            for j in range(n_sample)
        ],
        dims=["sample", "time"],
    )
    agg.record_batch(time, target_data, gen_data)
    if use_mock_distributed:
        world_size = 2
        with mock_distributed(world_size=world_size):
            result = agg._get_gathered_means()
            assert result is not None
            target, gen = result
            combined = agg.get_dataset()
    else:
        world_size = 1
        result = agg._get_gathered_means()
        assert result is not None
        target, gen = result
        combined = agg.get_dataset()
    for dataset in (target, gen, combined):
        assert set(dataset.dims).issuperset({"sample", "year"})
        assert list(dataset.year.values) == [2000, 2001, 2002]
        assert dataset.sizes["sample"] == n_sample * world_size
    assert set(combined.coords["source"].values) == set(["target", "prediction"])



import dataclasses
from typing import Dict, Iterable, List, Mapping, Optional, Union

import netCDF4
import numpy as np
import torch
import torch.jit

from fme.core.device import move_tensordict_to_device
from fme.core.typing_ import TensorDict, TensorMapping


@dataclasses.dataclass
class NormalizationConfig:
    """
    Configuration for normalizing data.

    Either global_means_path and global_stds_path or explicit means and stds
    must be provided.

    Parameters:
        global_means_path: Path to a netCDF file containing global means.
        global_stds_path: Path to a netCDF file containing global stds.
        means: Mapping from variable names to means.
        stds: Mapping from variable names to stds.
    """

    global_means_path: Optional[str] = None
    global_stds_path: Optional[str] = None
    means: Mapping[str, float] = dataclasses.field(default_factory=dict)
    stds: Mapping[str, float] = dataclasses.field(default_factory=dict)

    def __post_init__(self):
        using_path = (
            self.global_means_path is not None and self.global_stds_path is not None
        )
        using_explicit = len(self.means) > 0 and len(self.stds) > 0
        if using_path and using_explicit:
            raise ValueError(
                "Cannot use both global_means_path and global_stds_path "
                "and explicit means and stds."
            )
        if not (using_path or using_explicit):
            raise ValueError(
                "Must use either global_means_path and global_stds_path "
                "or explicit means and stds."
            )

    def build(self, names: List[str]):
        using_path = (
            self.global_means_path is not None and self.global_stds_path is not None
        )
        if using_path:
            return get_normalizer(
                global_means_path=self.global_means_path,
                global_stds_path=self.global_stds_path,
                names=names,
            )
        else:
            means = {k: torch.tensor(self.means[k]) for k in names}
            stds = {k: torch.tensor(self.stds[k]) for k in names}
            return StandardNormalizer(means=means, stds=stds)


class StandardNormalizer:
    """
    Responsible for normalizing tensors.
    """

    def __init__(
        self,
        means: TensorDict,
        stds: TensorDict,
    ):
        self.means = move_tensordict_to_device(means)
        self.stds = move_tensordict_to_device(stds)
        self._names = set(means).intersection(stds)

    def normalize(self, tensors: TensorMapping) -> TensorDict:
        filtered_tensors = {k: v for k, v in tensors.items() if k in self._names}
        return _normalize(filtered_tensors, means=self.means, stds=self.stds)

    def denormalize(self, tensors: TensorMapping) -> TensorDict:
        filtered_tensors = {k: v for k, v in tensors.items() if k in self._names}
        return _denormalize(filtered_tensors, means=self.means, stds=self.stds)

    def get_state(self):
        """
        Returns state as a serializable data structure.
        """
        return {
            "means": {k: float(v.cpu().numpy().item()) for k, v in self.means.items()},
            "stds": {k: float(v.cpu().numpy().item()) for k, v in self.stds.items()},
        }

    @classmethod
    def from_state(cls, state) -> "StandardNormalizer":
        """
        Loads state from a serializable data structure.
        """
        means = {
            k: torch.tensor(v, dtype=torch.float) for k, v in state["means"].items()
        }
        stds = {k: torch.tensor(v, dtype=torch.float) for k, v in state["stds"].items()}
        return cls(means=means, stds=stds)


@torch.jit.script
def _normalize(
    tensors: TensorDict,
    means: TensorDict,
    stds: TensorDict,
) -> TensorDict:
    return {k: (t - means[k]) / stds[k] for k, t in tensors.items()}


@torch.jit.script
def _denormalize(
    tensors: TensorDict,
    means: TensorDict,
    stds: TensorDict,
) -> TensorDict:
    return {k: t * stds[k] + means[k] for k, t in tensors.items()}


def get_normalizer(
    global_means_path, global_stds_path, names: List[str]
) -> StandardNormalizer:
    means = load_dict_from_netcdf(
        global_means_path, names, defaults={"x": 0.0, "y": 0.0, "z": 0.0}
    )
    means = {k: torch.as_tensor(v, dtype=torch.float) for k, v in means.items()}
    stds = load_dict_from_netcdf(
        global_stds_path, names, defaults={"x": 1.0, "y": 1.0, "z": 1.0}
    )
    stds = {k: torch.as_tensor(v, dtype=torch.float) for k, v in stds.items()}
    return StandardNormalizer(means=means, stds=stds)


def load_dict_from_netcdf(
    path: str, names: Iterable[str], defaults: Mapping[str, Union[float, np.ndarray]]
) -> Dict[str, Union[float, np.ndarray]]:
    """
    Load a dictionary of variables from a netCDF file.

    Args:
        path: Path to the netCDF file.
        names: List of variable names to load.
        defaults: Dictionary of default values for each variable, if not found
            in the netCDF file.
    """
    ds = netCDF4.Dataset(path)
    ds.set_auto_mask(False)
    result = {}
    for c in names:
        if c in ds.variables:
            result[c] = ds.variables[c][:]
        elif c in defaults:
            result[c] = defaults[c]
        else:
            raise ValueError(f"Variable {c} not found in {path}")
    ds.close()
    return result



import argparse
import dataclasses
import logging
import os
from typing import Callable, Optional

import dacite
import torch
import yaml

import fme
import fme.core.logging_utils as logging_utils
from fme.ace.aggregator.inference import InferenceEvaluatorAggregatorConfig
from fme.ace.data_loading.batch_data import BatchData, InferenceGriddedData
from fme.ace.data_loading.getters import get_inference_data
from fme.ace.data_loading.inference import InferenceDataLoaderConfig
from fme.ace.inference.data_writer import DataWriterConfig, PairedDataWriter
from fme.ace.inference.data_writer.time_coarsen import TimeCoarsenConfig
from fme.ace.inference.loop import (
    DeriverABC,
    run_dataset_comparison,
    write_reduced_metrics,
)
from fme.ace.stepper import SingleModuleStepper, SingleModuleStepperConfig
from fme.core.dicts import to_flat_dict
from fme.core.generics.inference import get_record_to_wandb, run_inference
from fme.core.logging_utils import LoggingConfig
from fme.core.ocean import OceanConfig
from fme.core.timing import GlobalTimer
from fme.core.typing_ import TensorDict, TensorMapping


def load_stepper_config(
    checkpoint_file: str, ocean_config: Optional[OceanConfig]
) -> SingleModuleStepperConfig:
    checkpoint = torch.load(checkpoint_file, map_location=fme.get_device())
    config = SingleModuleStepperConfig.from_state(checkpoint["stepper"]["config"])
    if ocean_config is not None:
        logging.info(
            "Overriding training ocean configuration with the inference ocean config."
        )
        config.ocean = ocean_config
    return config


def load_stepper(
    checkpoint_file: str,
    ocean_config: Optional[OceanConfig] = None,
) -> SingleModuleStepper:
    checkpoint = torch.load(checkpoint_file, map_location=fme.get_device())
    stepper = SingleModuleStepper.from_state(checkpoint["stepper"])
    if ocean_config is not None:
        logging.info(
            "Overriding training ocean configuration with the inference ocean config."
        )
        new_ocean = ocean_config.build(
            stepper.in_names, stepper.out_names, stepper.timestep
        )
        stepper.replace_ocean(new_ocean)
    return stepper


def validate_time_coarsen_config(
    config: TimeCoarsenConfig, forward_steps_in_memory: int, n_forward_steps: int
):
    coarsen_factor = config.coarsen_factor
    if forward_steps_in_memory % coarsen_factor != 0:
        raise ValueError(
            "forward_steps_in_memory must be divisible by "
            f"time_coarsen.coarsen_factor. Got {forward_steps_in_memory} "
            f"and {coarsen_factor}."
        )
    if n_forward_steps % coarsen_factor != 0:
        raise ValueError(
            "n_forward_steps must be divisible by "
            f"time_coarsen.coarsen_factor. Got {n_forward_steps} "
            f"and {coarsen_factor}."
        )


@dataclasses.dataclass
class InferenceEvaluatorConfig:
    """
    Configuration for running inference including comparison to reference data.

    Parameters:
        experiment_dir: Directory to save results to.
        n_forward_steps: Number of steps to run the model forward for.
        checkpoint_path: Path to stepper checkpoint to load.
        logging: configuration for logging.
        loader: Configuration for data to be used as initial conditions, forcing, and
            target in inference.
        prediction_loader: Configuration for prediction data to evaluate. If given,
            model evaluation will not run, and instead predictions will be evaluated.
            Model checkpoint will still be used to determine inputs and outputs.
        forward_steps_in_memory: Number of forward steps to complete in memory
            at a time, will load one more step for initial condition.
        data_writer: Configuration for data writers.
        aggregator: Configuration for inference evaluator aggregator.
        ocean: Ocean configuration for running inference with a
            different one than what is used in training.
    """

    experiment_dir: str
    n_forward_steps: int
    checkpoint_path: str
    logging: LoggingConfig
    loader: InferenceDataLoaderConfig
    prediction_loader: Optional[InferenceDataLoaderConfig] = None
    forward_steps_in_memory: int = 1
    data_writer: DataWriterConfig = dataclasses.field(
        default_factory=lambda: DataWriterConfig()
    )
    aggregator: InferenceEvaluatorAggregatorConfig = dataclasses.field(
        default_factory=lambda: InferenceEvaluatorAggregatorConfig()
    )
    ocean: Optional[OceanConfig] = None

    def __post_init__(self):
        if self.data_writer.time_coarsen is not None:
            validate_time_coarsen_config(
                self.data_writer.time_coarsen,
                self.forward_steps_in_memory,
                self.n_forward_steps,
            )

    def configure_logging(self, log_filename: str):
        self.logging.configure_logging(self.experiment_dir, log_filename)

    def configure_wandb(self, env_vars: Optional[dict] = None, **kwargs):
        config = to_flat_dict(dataclasses.asdict(self))
        self.logging.configure_wandb(
            config=config, env_vars=env_vars, resume=False, **kwargs
        )

    def clean_wandb(self):
        self.logging.clean_wandb(self.experiment_dir)

    def load_stepper(self) -> SingleModuleStepper:
        logging.info(f"Loading trained model checkpoint from {self.checkpoint_path}")
        stepper = load_stepper(self.checkpoint_path, ocean_config=self.ocean)
        return stepper

    def load_stepper_config(self) -> SingleModuleStepperConfig:
        logging.info(f"Loading trained model checkpoint from {self.checkpoint_path}")
        return load_stepper_config(self.checkpoint_path, ocean_config=self.ocean)

    def get_data_writer(self, data: InferenceGriddedData) -> PairedDataWriter:
        return self.data_writer.build_paired(
            experiment_dir=self.experiment_dir,
            n_initial_conditions=self.loader.n_initial_conditions,
            n_timesteps=self.n_forward_steps,
            timestep=data.timestep,
            variable_metadata=data.variable_metadata,
            coords=data.coords,
        )


def main(yaml_config: str):
    with open(yaml_config, "r") as f:
        data = yaml.safe_load(f)
    config = dacite.from_dict(
        data_class=InferenceEvaluatorConfig,
        data=data,
        config=dacite.Config(strict=True),
    )
    if not os.path.isdir(config.experiment_dir):
        os.makedirs(config.experiment_dir, exist_ok=True)
    with open(os.path.join(config.experiment_dir, "config.yaml"), "w") as f:
        yaml.dump(data, f, default_flow_style=False, sort_keys=False)
    with GlobalTimer():
        return run_evaluator_from_config(config)


class _Deriver(DeriverABC):
    """
    DeriverABC implementation for dataset comparison.
    """

    def __init__(
        self,
        n_ic_timesteps: int,
        derive_func: Callable[[TensorMapping, TensorMapping], TensorDict],
    ):
        self._n_ic_timesteps = n_ic_timesteps
        self._derive_func = derive_func

    @property
    def n_ic_timesteps(self) -> int:
        return self._n_ic_timesteps

    def get_forward_data(
        self, data: BatchData, compute_derived_variables: bool = False
    ) -> BatchData:
        if compute_derived_variables:
            timer = GlobalTimer.get_instance()
            with timer.context("compute_derived_variables"):
                data = data.compute_derived_variables(
                    derive_func=self._derive_func,
                    forcing_data=data,
                )
        return data.remove_initial_condition(self._n_ic_timesteps)


def run_evaluator_from_config(config: InferenceEvaluatorConfig):
    timer = GlobalTimer.get_instance()
    timer.start_outer("inference")
    timer.start("initialization")

    if not os.path.isdir(config.experiment_dir):
        os.makedirs(config.experiment_dir, exist_ok=True)
    config.configure_logging(log_filename="inference_out.log")
    env_vars = logging_utils.retrieve_env_vars()
    beaker_url = logging_utils.log_beaker_url()
    config.configure_wandb(env_vars=env_vars, notes=beaker_url)

    torch.backends.cudnn.benchmark = True

    logging_utils.log_versions()
    logging.info(f"Current device is {fme.get_device()}")

    stepper_config = config.load_stepper_config()
    logging.info("Loading inference data")
    window_requirements = stepper_config.get_evaluation_window_data_requirements(
        n_forward_steps=config.forward_steps_in_memory
    )
    initial_condition_requirements = (
        stepper_config.get_prognostic_state_data_requirements()
    )
    data = get_inference_data(
        config=config.loader,
        total_forward_steps=config.n_forward_steps,
        window_requirements=window_requirements,
        initial_condition=initial_condition_requirements,
    )

    stepper = config.load_stepper()
    if stepper.timestep != data.timestep:
        raise ValueError(
            f"Timestep of the loaded stepper, {stepper.timestep}, does not "
            f"match that of the forcing data, {data.timestep}."
        )

    aggregator_config: InferenceEvaluatorAggregatorConfig = config.aggregator
    for batch in data.loader:
        initial_time = batch.time.isel(time=0)
        break
    aggregator = aggregator_config.build(
        vertical_coordinate=data.vertical_coordinate,
        horizontal_coordinates=data.horizontal_coordinates,
        timestep=data.timestep,
        record_step_20=config.n_forward_steps >= 20,
        n_timesteps=config.n_forward_steps + stepper_config.n_ic_timesteps,
        variable_metadata=data.variable_metadata,
        initial_time=initial_time,
        channel_mean_names=stepper.out_names,
        normalize=stepper.normalizer.normalize,
    )

    writer = config.get_data_writer(data)

    timer.stop()
    logging.info("Starting inference")
    record_logs = get_record_to_wandb(label="inference")
    if config.prediction_loader is not None:
        prediction_data = get_inference_data(
            config.prediction_loader,
            total_forward_steps=config.n_forward_steps,
            window_requirements=window_requirements,
            initial_condition=initial_condition_requirements,
        )
        deriver = _Deriver(
            n_ic_timesteps=stepper_config.n_ic_timesteps,
            derive_func=stepper.derive_func,
        )
        run_dataset_comparison(
            aggregator=aggregator,
            prediction_data=prediction_data,
            target_data=data,
            deriver=deriver,
            writer=writer,
            record_logs=record_logs,
        )
    else:
        run_inference(
            predict=stepper.predict_paired,
            data=data,
            aggregator=aggregator,
            writer=writer,
            record_logs=record_logs,
        )

    timer.start("final_writer_flush")
    logging.info("Starting final flush of data writer")
    writer.flush()
    logging.info("Writing reduced metrics to disk in netcdf format.")
    write_reduced_metrics(
        aggregator,
        data.coords,
        config.experiment_dir,
        excluded=[
            "video",
        ],
    )
    timer.stop()

    timer.stop_outer("inference")
    total_steps = config.n_forward_steps * config.loader.n_initial_conditions
    inference_duration = timer.get_duration("inference")
    wandb_logging_duration = timer.get_duration("wandb_logging")
    total_steps_per_second = total_steps / (inference_duration - wandb_logging_duration)
    timer.log_durations()
    logging.info(
        "Total steps per second (ignoring wandb logging): "
        f"{total_steps_per_second:.2f} steps/second"
    )

    summary_logs = {
        "total_steps_per_second": total_steps_per_second,
        **timer.get_durations(),
        **aggregator.get_summary_logs(),
    }
    record_logs([summary_logs])
    config.clean_wandb()


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("yaml_config", type=str)

    args = parser.parse_args()

    main(
        yaml_config=args.yaml_config,
    )



import contextlib
import dataclasses
import itertools
from typing import Any, Iterable, Literal, Mapping, Optional

import torch
import torch.cuda.amp as amp
from torch import nn

from fme.core.generics.optimization import OptimizationABC
from fme.core.scheduler import SchedulerConfig


class Optimization(OptimizationABC):
    def __init__(
        self,
        parameters: Iterable[torch.nn.Parameter],
        optimizer_type: Literal["Adam", "FusedAdam"],
        lr: float,
        max_epochs: int,
        scheduler: SchedulerConfig,
        enable_automatic_mixed_precision: bool,
        kwargs: Mapping[str, Any],
    ):
        if optimizer_type == "FusedAdam":
            self.optimizer = torch.optim.AdamW(parameters, lr=lr, fused=True, **kwargs)
        elif optimizer_type == "Adam":
            self.optimizer = torch.optim.Adam(parameters, lr=lr, **kwargs)
        else:
            raise ValueError(f"Unknown optimizer type: {optimizer_type}")

        if enable_automatic_mixed_precision:
            self.gscaler: Optional[amp.GradScaler] = amp.GradScaler()
        else:
            self.gscaler = None
        self.scheduler = scheduler.build(self.optimizer, max_epochs)

    @contextlib.contextmanager
    def autocast(self):
        enabled = self.gscaler is not None
        dtype = torch.bfloat16 if enabled else None
        with amp.autocast(enabled=enabled, dtype=dtype):
            yield

    @property
    def learning_rate(self) -> float:
        return self.optimizer.param_groups[0]["lr"]

    def set_mode(self, module: nn.Module):
        """
        Sets the mode of the module to train.
        """
        module.train()

    def step_scheduler(self, valid_loss: float):
        """
        Step the scheduler.

        Args:
            valid_loss: The validation loss. Used in schedulers which change the
                learning rate based on whether the validation loss is decreasing.
        """
        if self.scheduler is not None:
            try:
                self.scheduler.step(metrics=valid_loss)
            except TypeError:
                self.scheduler.step()

    def step_weights(self, loss: torch.Tensor):
        self._validate_loss(loss)

        if self.gscaler is not None:
            self.gscaler.scale(loss).backward()
            self.gscaler.step(self.optimizer)
        else:
            loss.backward()
            self.optimizer.step()
        self.optimizer.zero_grad()

        if self.gscaler is not None:
            self.gscaler.update()

    def get_state(self):
        """
        Returns state as a serializable data structure.
        """
        state = {
            "optimizer_state_dict": self.optimizer.state_dict(),
            "scheduler_state_dict": (
                self.scheduler.state_dict() if self.scheduler is not None else None
            ),
            "gscaler_state_dict": (
                self.gscaler.state_dict() if self.gscaler is not None else None
            ),
        }
        return state

    def load_state(self, state):
        """
        Loads state from a serializable data structure.
        """
        self.optimizer.load_state_dict(state["optimizer_state_dict"])
        if self.scheduler is not None:
            self.scheduler.load_state_dict(state["scheduler_state_dict"])
        if self.gscaler is not None:
            self.gscaler.load_state_dict(state["gscaler_state_dict"])

    def _validate_loss(self, loss: torch.Tensor):
        with torch.no_grad():
            if torch.isnan(loss):
                raise ValueError("Loss is NaN-valued during training.")


@dataclasses.dataclass
class OptimizationConfig:
    """
    Configuration for optimization.

    Parameters:
        optimizer_type: The type of optimizer to use.
        lr: The learning rate.
        kwargs: Additional keyword arguments to pass to the optimizer.
        enable_automatic_mixed_precision: Whether to use automatic mixed
            precision.
        scheduler: The type of scheduler to use. If none is given, no scheduler
            will be used.
    """

    optimizer_type: Literal["Adam", "FusedAdam"] = "Adam"
    lr: float = 0.001
    kwargs: Mapping[str, Any] = dataclasses.field(default_factory=dict)
    enable_automatic_mixed_precision: bool = False
    scheduler: SchedulerConfig = dataclasses.field(
        default_factory=lambda: SchedulerConfig()
    )

    def build(self, modules: torch.nn.ModuleList, max_epochs: int) -> Optimization:
        parameters = itertools.chain(*[module.parameters() for module in modules])
        return Optimization(
            parameters=parameters,
            optimizer_type=self.optimizer_type,
            lr=self.lr,
            max_epochs=max_epochs,
            scheduler=self.scheduler,
            enable_automatic_mixed_precision=self.enable_automatic_mixed_precision,
            kwargs=self.kwargs,
        )

    def get_state(self) -> Mapping[str, Any]:
        return dataclasses.asdict(self)

    @classmethod
    def from_state(cls, state: Mapping[str, Any]) -> "OptimizationConfig":
        return cls(**state)


class NullOptimization(OptimizationABC):
    @contextlib.contextmanager
    def autocast(self):
        yield

    @property
    def learning_rate(self) -> float:
        return float("nan")

    def step_scheduler(self, valid_loss: float):
        return

    def step_weights(self, loss: torch.Tensor):
        return

    def get_state(self):
        return {}

    def load_state(self, state):
        return

    def set_mode(self, module: nn.Module):
        """
        Sets the mode of the module to eval.
        """
        module.eval()



import abc
from typing import Any, Dict, List, Type, TypeVar

import torch

from fme.core import metrics
from fme.core.device import get_device


class GriddedOperations(abc.ABC):
    @abc.abstractmethod
    def area_weighted_mean(
        self, data: torch.Tensor, keepdim: bool = False
    ) -> torch.Tensor: ...

    def area_weighted_mean_bias(
        self, truth: torch.Tensor, predicted: torch.Tensor
    ) -> torch.Tensor:
        return self.area_weighted_mean(predicted - truth)

    def area_weighted_rmse(
        self, truth: torch.Tensor, predicted: torch.Tensor
    ) -> torch.Tensor:
        return torch.sqrt(self.area_weighted_mean((predicted - truth) ** 2))

    def area_weighted_std(self, data: torch.Tensor, keepdim: bool = False):
        return self.area_weighted_mean(
            (data - self.area_weighted_mean(data, keepdim=True)) ** 2,
            keepdim=keepdim,
        ).sqrt()

    @abc.abstractmethod
    def area_weighted_gradient_magnitude_percent_diff(
        self, truth: torch.Tensor, predicted: torch.Tensor
    ): ...

    def to_state(self) -> Dict[str, Any]:
        return {
            "type": self.__class__.__name__,
            "state": self.get_initialization_kwargs(),
        }

    @abc.abstractmethod
    def get_initialization_kwargs(self) -> Dict[str, Any]:
        """
        Get the keyword arguments needed to initialize the instance.
        """
        ...

    @classmethod
    def from_state(cls, state: Dict[str, Any]) -> "GriddedOperations":
        """
        Given a dictionary with a "type" key and a "state" key, return
        the GriddedOperations it describes.

        The "type" key should be the name of a subclass of GriddedOperations,
        and the "state" key should be a dictionary specific to
        that subclass.

        Args:
            state: A dictionary with a "type" key and a "state" key.

        Returns:
            An instance of the subclass.
        """
        if cls is not GriddedOperations:
            raise RuntimeError(
                "This method should be called on GriddedOperations, "
                "not on its subclasses."
            )
        subclasses = get_all_subclasses(cls)
        for subclass in subclasses:
            if subclass.__name__ == state["type"]:
                return subclass(**state["state"])
        raise ValueError(
            f"Unknown subclass type: {state['type']}, "
            f"available: {[s.__name__ for s in subclasses]}"
        )


T = TypeVar("T")


def get_all_subclasses(cls: Type[T]) -> List[Type[T]]:
    """
    Gets all subclasses of a given class, including their subclasses etc.
    """
    all_subclasses = []
    for subclass in cls.__subclasses__():
        all_subclasses.append(subclass)
        all_subclasses.extend(get_all_subclasses(subclass))
    return all_subclasses


class LatLonOperations(GriddedOperations):
    HORIZONTAL_DIMS = (-2, -1)

    def __init__(self, area_weights: torch.Tensor):
        self._device_area = area_weights.to(get_device())
        self._cpu_area = area_weights.to("cpu")

    def area_weighted_mean(
        self, data: torch.Tensor, keepdim: bool = False
    ) -> torch.Tensor:
        if data.device.type == "cpu":
            area_weights = self._cpu_area
        else:
            area_weights = self._device_area
        return metrics.weighted_mean(
            data, area_weights, dim=self.HORIZONTAL_DIMS, keepdim=keepdim
        )

    def area_weighted_gradient_magnitude_percent_diff(
        self, truth: torch.Tensor, predicted: torch.Tensor
    ):
        if predicted.device.type == "cpu":
            area_weights = self._cpu_area
        else:
            area_weights = self._device_area
        return metrics.gradient_magnitude_percent_diff(
            truth, predicted, weights=area_weights, dim=self.HORIZONTAL_DIMS
        )

    def get_initialization_kwargs(self) -> Dict[str, Any]:
        return {"area_weights": self._cpu_area}


class HEALPixOperations(GriddedOperations):
    HORIZONTAL_DIMS = (-3, -2, -1)

    def area_weighted_mean(
        self, data: torch.Tensor, keepdim: bool = False
    ) -> torch.Tensor:
        return data.mean(dim=self.HORIZONTAL_DIMS, keepdim=keepdim)

    def area_weighted_gradient_magnitude_percent_diff(
        self, truth: torch.Tensor, predicted: torch.Tensor
    ):
        return metrics.gradient_magnitude_percent_diff(
            truth, predicted, weights=None, dim=self.HORIZONTAL_DIMS
        )

    def get_initialization_kwargs(self) -> Dict[str, Any]:
        return {}



import dataclasses
from typing import Optional

import torch

from fme.core.stacker import Stacker, unstack
from fme.core.typing_ import TensorDict, TensorMapping


def replace_on_mask(
    original: torch.Tensor,
    replacement: torch.Tensor,
    mask: torch.Tensor,
    mask_value: int,
):
    """Replace original with replacement in masked regions.

    Args:
        original: The original data tensor.
        replacement: The replacement data tensor.
        mask: The mask tensor.
        mask_value: The value of the mask variable in the region to be replaced.
    """
    rounded_mask = torch.round(mask).to(int)
    return torch.where(
        condition=rounded_mask == mask_value,
        input=replacement,
        other=original,
    )


@dataclasses.dataclass
class MaskingConfig:
    """
    Configuration for applying masking to the generated output.

    Parameters:
        mask_name: The standard name of the mask. May be a prefix (for a 3D masking
            variable) or a full name (for a 2D masking variable).
        mask_value: Value of the mask variable in masked regions. Either 0 or 1.
        fill_value: The constant fill value to use outside of masked regions.
        surface_mask_name: (optional) The full name of the surface mask. Only required
            when mask_name is a prefix and separate 2D surface masking is desired.
    """

    mask_name: str
    mask_value: int
    fill_value: float = 0.0
    surface_mask_name: Optional[str] = None

    def __post_init__(self):
        if self.mask_value not in [0, 1]:
            raise ValueError(
                "mask_value must be either 0 or 1, but got " f"{self.mask_value}"
            )

    def build(self):
        return Masking(
            mask_name=self.mask_name,
            mask_value=self.mask_value,
            fill_value=self.fill_value,
            surface_mask_name=self.surface_mask_name,
        )


class Masking:
    """Replace masked regions with a fill value."""

    def __init__(
        self,
        mask_name: str,
        mask_value: int,
        fill_value: float,
        surface_mask_name: Optional[str] = None,
    ):
        self.mask_name = mask_name
        self.mask_value = mask_value
        self.fill_value = fill_value
        self.surface_mask_name = surface_mask_name
        mask_map = {self.mask_name: [self.mask_name]}
        if self.surface_mask_name is not None:
            mask_map[self.surface_mask_name] = [self.surface_mask_name]

        self.mask_stacker = Stacker(mask_map)

    def __call__(
        self,
        stacker: Stacker,
        data: TensorMapping,
        mask_data: TensorMapping,
    ) -> TensorDict:
        """
        Apply masking to the data for standard names recognized by a stacker.

        Args:
            stacker: A Stacker for variables to mask in data.
            data: The data to mask.
            mask_data: The mask data.

        """
        mask = self.mask_stacker(self.mask_name, mask_data)
        if self.surface_mask_name is not None:
            surface_mask = self.mask_stacker(self.surface_mask_name, mask_data)
        else:
            surface_mask = None
        data_: TensorDict = {**data}
        for name in stacker.standard_names:
            stacked = stacker(name, data)
            if stacked.size(-1) > 1:  # 3D masking
                mask_ = mask
            elif surface_mask is not None:  # 2D masking with surface mask
                mask_ = surface_mask
            elif mask.size(-1) == 1:  # 2D masking with mask
                mask_ = mask
            else:
                raise RuntimeError(
                    "Masking surface_mask_name is None but the input Stacker "
                    f"includes the 2D standard name {name}."
                )
            fill = torch.full_like(stacked, self.fill_value)
            masked = replace_on_mask(
                original=stacked,
                replacement=fill,
                mask=mask_,
                mask_value=self.mask_value,
            )
            level_names = stacker.get_all_level_names(name, data)
            data_.update(unstack(masked, level_names, dim=-1))
        return data_



import copy
import datetime
from typing import List

import numpy as np
import pytest
import torch
import xarray as xr

from fme.core.coordinates import (
    DimSize,
    HEALPixCoordinates,
    HorizontalCoordinates,
    LatLonCoordinates,
)

from .utils import (
    _get_indexers,
    as_broadcasted_tensor,
    decode_timestep,
    encode_timestep,
    infer_horizontal_dimension_names,
)

LON_DIM = "lon"
LAT_DIM = "lat"
TIME_DIM = "time"
FULL_NAME = "full"
CONSTANT_NAME = "constant"
SLICE_NONE = slice(None)


def get_sizes(
    spatial_dims: HorizontalCoordinates = LatLonCoordinates(
        lon=torch.Tensor(np.arange(6)),
        lat=torch.Tensor(np.arange(12)),
        loaded_lat_name=LAT_DIM,
        loaded_lon_name=LON_DIM,
    ),
):
    spatial_sizes: List[DimSize] = copy.deepcopy(spatial_dims.loaded_default_sizes)
    spatial_sizes.append(DimSize(TIME_DIM, 3))
    return spatial_sizes


def create_reference_dataset(
    spatial_dims: HorizontalCoordinates = LatLonCoordinates(
        lon=torch.Tensor(np.arange(6)),
        lat=torch.Tensor(np.arange(12)),
        loaded_lat_name=LAT_DIM,
        loaded_lon_name=LON_DIM,
    ),
):
    dims = [TIME_DIM] + spatial_dims.loaded_dims
    dim_sizes = get_sizes(spatial_dims=spatial_dims)
    shape = tuple(dim_size.size for dim_size in dim_sizes)
    data = np.arange(np.prod(shape)).reshape(shape)
    coords = [np.arange(size) for size in shape]
    full = xr.DataArray(data, dims=dims, coords=coords, name=FULL_NAME)
    constant = full.isel({TIME_DIM: 0}).drop_vars(TIME_DIM).rename(CONSTANT_NAME)
    return xr.merge([full, constant])


@pytest.mark.parametrize(
    ("lon_dim", "lat_dim", "warns"),
    [
        ("grid_xt", "grid_yt", False),
        ("lon", "lat", False),
        ("longitude", "latitude", False),
        ("foo", "bar", True),
    ],
)
def test_infer_horizontal_dimension_names(lon_dim, lat_dim, warns):
    spatial_dims = LatLonCoordinates(
        lon=torch.Tensor(np.arange(6)),
        lat=torch.Tensor(np.arange(12)),
        loaded_lat_name=lat_dim,
        loaded_lon_name=lon_dim,
    )
    ds = create_reference_dataset(spatial_dims=spatial_dims)
    expected = [lon_dim, lat_dim]
    for dim in expected:
        assert dim in ds.dims
    if warns:
        with pytest.warns(UserWarning, match="Familiar"):
            infer_horizontal_dimension_names(ds)
    else:
        result = infer_horizontal_dimension_names(ds)
        assert result == expected


def test_infer_horizontal_dimension_names_healpix():
    # create a dummy HEALPixCoordinates to test other data format
    hpx_coords = HEALPixCoordinates(
        face=torch.Tensor(np.arange(12)),
        width=torch.Tensor(np.arange(64)),
        height=torch.Tensor(np.arange(64)),
    )
    ds = create_reference_dataset(spatial_dims=hpx_coords)
    expected = hpx_coords.loaded_dims
    result = infer_horizontal_dimension_names(ds)
    assert result == expected


@pytest.mark.parametrize(
    ["coordinate_type", "coord_sizes"],
    [
        pytest.param(LatLonCoordinates, {"lat": 90, "lon": 180}),
        pytest.param(HEALPixCoordinates, {"face": 12, "height": 64, "width": 64}),
    ],
)
def test_horizonal_dimension_sizes(coordinate_type, coord_sizes):
    coords = {name: torch.Tensor(np.arange(size)) for name, size in coord_sizes.items()}
    horizontal_coords = coordinate_type(**coords)
    for name, size in coord_sizes.items():
        assert len(horizontal_coords.coords[name]) == size


def test_infer_horizontal_dimension_names_error():
    spatial_dims = LatLonCoordinates(
        lon=torch.Tensor(np.arange(6)),
        lat=torch.Tensor(np.arange(12)),
        loaded_lat_name="foo",
        loaded_lon_name="bar",
    )
    ds = create_reference_dataset(spatial_dims=spatial_dims)
    ds = ds.isel(time=0)
    print(ds)
    with pytest.raises(ValueError, match="Could not identify"):
        infer_horizontal_dimension_names(ds)


@pytest.mark.parametrize(
    ("variable_dims", "expected"),
    [
        ((TIME_DIM, LAT_DIM, LON_DIM), (SLICE_NONE, SLICE_NONE, SLICE_NONE)),
        ((LAT_DIM, LON_DIM), (None, SLICE_NONE, SLICE_NONE)),
        ((TIME_DIM,), (SLICE_NONE, None, None)),
        ((LAT_DIM,), (None, SLICE_NONE, None)),
        ((LON_DIM,), (None, None, SLICE_NONE)),
    ],
    ids=lambda x: f"{x}",
)
def test__get_indexers(variable_dims, expected):
    dim_sizes = get_sizes()
    shape = tuple(
        dim_size.size for dim_size in dim_sizes if dim_size.name in variable_dims
    )
    variable = xr.Variable(variable_dims, np.zeros(shape))
    dims = (TIME_DIM, LAT_DIM, LON_DIM)
    result = _get_indexers(variable, dims)
    assert result == expected


@pytest.mark.parametrize(
    "variable_dims",
    [
        (TIME_DIM, LAT_DIM, LON_DIM),
        (LAT_DIM, LON_DIM),
        (TIME_DIM,),
        (LAT_DIM,),
        (LON_DIM,),
    ],
)
def test_as_broadcasted_tensor(variable_dims):
    ds = create_reference_dataset()

    variable_shape = tuple(ds.sizes[dim] for dim in variable_dims)
    variable = xr.Variable(variable_dims, np.zeros(variable_shape))
    da = xr.DataArray(variable)

    dims = ds[FULL_NAME].dims
    shape = (ds.sizes[TIME_DIM], ds.sizes[LAT_DIM], ds.sizes[LON_DIM])
    result = as_broadcasted_tensor(variable, dims, shape)

    xarray_broadcast = da.broadcast_like(ds[FULL_NAME]).transpose(*dims)

    # Suppress read-only warning when casting to a tensor; note this is only
    # an issue following an xarray broadcast, which is used for convenience
    # in this test. The internals of the data loader convert the array to a
    # tensor prior to the broadcast, which avoids this issue.
    xarray_broadcast = xarray_broadcast.copy()

    expected = torch.as_tensor(xarray_broadcast.values)

    assert torch.equal(result, expected)


def test_encode_decode_timestep_roundtrip():
    timestep = datetime.timedelta(days=1, microseconds=1)
    encoded = encode_timestep(timestep)
    roundtripped = decode_timestep(encoded)
    assert roundtripped == timestep



"""
This file contains code for computing metrics of single variables on batches of data,
and aggregating them into a single metric value. The functions here mainly exist
to turn metric functions that may have different APIs into a common API,
so that they can be iterated over and called in the same way in a loop.
"""

from typing import Protocol

import torch


class AreaWeightedFunction(Protocol):
    """
    A function that computes a metric on the true and predicted values,
    weighted by area.
    """

    def __call__(
        self,
        truth: torch.Tensor,
        predicted: torch.Tensor,
    ) -> torch.Tensor: ...


class ReducedMetric(Protocol):
    """Used to record a metric value on batches of data (potentially out-of-memory)
    and then get the total metric at the end.
    """

    def record(self, target: torch.Tensor, gen: torch.Tensor):
        """
        Update metric for a batch of data.
        """
        ...

    def get(self) -> torch.Tensor:
        """
        Get the total metric value, not divided by number of recorded batches.
        """
        ...


class AreaWeightedReducedMetric:
    """
    A wrapper around an area-weighted metric function.
    """

    def __init__(
        self,
        device: torch.device,
        compute_metric: AreaWeightedFunction,
    ):
        self._compute_metric = compute_metric
        self._total = None
        self._device = device

    def record(self, target: torch.Tensor, gen: torch.Tensor):
        """Add a batch of data to the metric.

        Args:
            target: Target data. Should have shape [batch, time, height, width].
            gen: Generated data. Should have shape [batch, time, height, width].
        """
        new_value = self._compute_metric(target, gen).mean(dim=0)
        if self._total is None:
            self._total = torch.zeros_like(new_value, device=self._device)
        self._total += new_value

    def get(self) -> torch.Tensor:
        """Returns the metric."""
        return self._total



# This module is derived from the train.py module in the following repository:
# https://github.com/NVlabs/FourCastNet. The corresponding license is
# provided below.

# BSD 3-Clause License
#
# Copyright (c) 2022, FourCastNet authors
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# The code was authored by the following people:
#
# Jaideep Pathak - NVIDIA Corporation
# Shashank Subramanian - NERSC, Lawrence Berkeley National Laboratory
# Peter Harrington - NERSC, Lawrence Berkeley National Laboratory
# Sanjeev Raja - NERSC, Lawrence Berkeley National Laboratory
# Ashesh Chattopadhyay - Rice University
# Morteza Mardani - NVIDIA Corporation
# Thorsten Kurth - NVIDIA Corporation
# David Hall - NVIDIA Corporation
# Zongyi Li - California Institute of Technology, NVIDIA Corporation
# Kamyar Azizzadenesheli - Purdue University
# Pedram Hassanzadeh - Rice University
# Karthik Kashinath - NVIDIA Corporation
# Animashree Anandkumar - California Institute of Technology, NVIDIA Corporation

import dataclasses
import logging
import os
from datetime import timedelta
from typing import Callable, Dict, Mapping, Optional, Sequence

import dacite
import dask
import torch
import xarray as xr
import yaml

import fme
import fme.core.logging_utils as logging_utils
from fme.ace.aggregator import OneStepAggregator, TrainAggregator
from fme.ace.aggregator.inference.main import (
    InferenceEvaluatorAggregator,
    InferenceEvaluatorAggregatorConfig,
)
from fme.ace.data_loading.batch_data import PairedData, PrognosticState
from fme.ace.stepper import TrainOutput
from fme.ace.train.train_config import TrainBuilders, TrainConfig
from fme.core.coordinates import HorizontalCoordinates, HybridSigmaPressureCoordinate
from fme.core.dataset.data_typing import VariableMetadata
from fme.core.dicts import to_flat_dict
from fme.core.distributed import Distributed
from fme.core.generics.trainer import AggregatorBuilderABC, TrainConfigProtocol, Trainer
from fme.core.gridded_ops import GriddedOperations
from fme.core.typing_ import TensorDict, TensorMapping

# dask used on individual workers to load batches
dask.config.set(scheduler="synchronous")


def build_trainer(builder: TrainBuilders, config: TrainConfig) -> "Trainer":
    # note for devs: you don't have to use this function to build a custom
    # trainer, you can build it however you like. This is here for convenience.
    train_data = builder.get_train_data()
    validation_data = builder.get_validation_data()
    inference_data = builder.get_evaluation_inference_data()

    for batch in train_data.loader:
        shapes = {k: v.shape for k, v in batch.data.items()}
        for value in shapes.values():
            img_shape = value[-2:]
            break
        break
    logging.info("Starting model initialization")
    stepper = builder.get_stepper(
        img_shape=img_shape,
        gridded_operations=train_data.gridded_operations,
        vertical_coordinate=train_data.vertical_coordinate,
        timestep=train_data.timestep,
    )
    end_of_batch_ops = builder.get_end_of_batch_ops(stepper.modules)

    for batch in inference_data.loader:
        initial_inference_times = batch.time.isel(time=0)
        break
    aggregator_builder = AggregatorBuilder(
        inference_config=config.inference_aggregator,
        gridded_operations=train_data.gridded_operations,
        vertical_coordinate=train_data.vertical_coordinate,
        horizontal_coordinates=train_data.horizontal_coordinates,
        timestep=train_data.timestep,
        initial_inference_time=initial_inference_times,
        record_step_20=config.inference_n_forward_steps >= 20,
        n_timesteps=config.inference_n_forward_steps + stepper.n_ic_timesteps,
        variable_metadata=train_data.variable_metadata,
        loss_scaling=stepper.effective_loss_scaling,
        channel_mean_names=stepper.out_names,
        normalize=stepper.normalizer.normalize,
    )
    do_gc_collect = fme.get_device() != torch.device("cpu")
    trainer_config: TrainConfigProtocol = config  # documenting trainer input type
    return Trainer(
        train_data=train_data,
        validation_data=validation_data,
        inference_data=inference_data,
        stepper=stepper,
        build_optimization=builder.get_optimization,
        build_ema=builder.get_ema,
        config=trainer_config,
        aggregator_builder=aggregator_builder,
        end_of_batch_callback=end_of_batch_ops,
        do_gc_collect=do_gc_collect,
    )


class AggregatorBuilder(
    AggregatorBuilderABC[PrognosticState, TrainOutput, PairedData],
):
    def __init__(
        self,
        inference_config: InferenceEvaluatorAggregatorConfig,
        gridded_operations: GriddedOperations,
        vertical_coordinate: HybridSigmaPressureCoordinate,
        horizontal_coordinates: HorizontalCoordinates,
        timestep: timedelta,
        initial_inference_time: xr.DataArray,
        record_step_20: bool,
        n_timesteps: int,
        normalize: Callable[[TensorMapping], TensorDict],
        variable_metadata: Optional[Mapping[str, VariableMetadata]] = None,
        loss_scaling: Optional[Dict[str, torch.Tensor]] = None,
        channel_mean_names: Optional[Sequence[str]] = None,
    ):
        self.inference_config = inference_config
        self.gridded_operations = gridded_operations
        self.vertical_coordinate = vertical_coordinate
        self.horizontal_coordinates = horizontal_coordinates
        self.timestep = timestep
        self.initial_inference_time = initial_inference_time
        self.record_step_20 = record_step_20
        self.n_timesteps = n_timesteps
        self.variable_metadata = variable_metadata
        self.loss_scaling = loss_scaling
        self.channel_mean_names = channel_mean_names
        self.normalize = normalize

    def get_train_aggregator(self) -> TrainAggregator:
        return TrainAggregator()

    def get_validation_aggregator(self) -> OneStepAggregator:
        return OneStepAggregator(
            gridded_operations=self.gridded_operations,
            variable_metadata=self.variable_metadata,
            loss_scaling=self.loss_scaling,
        )

    def get_inference_aggregator(
        self,
    ) -> InferenceEvaluatorAggregator:
        return self.inference_config.build(
            vertical_coordinate=self.vertical_coordinate,
            horizontal_coordinates=self.horizontal_coordinates,
            timestep=self.timestep,
            initial_time=self.initial_inference_time,
            record_step_20=self.record_step_20,
            n_timesteps=self.n_timesteps,
            variable_metadata=self.variable_metadata,
            channel_mean_names=self.channel_mean_names,
            normalize=self.normalize,
        )


def run_train_from_config(config: TrainConfig):
    run_train(TrainBuilders(config), config)


def run_train(builders: TrainBuilders, config: TrainConfig):
    dist = Distributed.get_instance()
    if fme.using_gpu():
        torch.backends.cudnn.benchmark = True
    if not os.path.isdir(config.experiment_dir):
        os.makedirs(config.experiment_dir, exist_ok=True)
    config.logging.configure_logging(config.experiment_dir, log_filename="out.log")
    env_vars = logging_utils.retrieve_env_vars()
    logging_utils.log_versions()
    beaker_url = logging_utils.log_beaker_url()
    config_as_dict = to_flat_dict(dataclasses.asdict(config))
    config.logging.configure_wandb(
        config=config_as_dict, env_vars=env_vars, resume=True, notes=beaker_url
    )
    trainer = build_trainer(builders, config)
    trainer.train()
    logging.info("DONE ---- rank %d" % dist.rank)


def main(yaml_config: str):
    with open(yaml_config, "r") as f:
        data = yaml.safe_load(f)
    train_config: TrainConfig = dacite.from_dict(
        data_class=TrainConfig,
        data=data,
        config=dacite.Config(strict=True),
    )
    if not os.path.isdir(train_config.experiment_dir):
        os.makedirs(train_config.experiment_dir, exist_ok=True)
    with open(os.path.join(train_config.experiment_dir, "config.yaml"), "w") as f:
        yaml.dump(data, f, default_flow_style=False, sort_keys=False)
    run_train_from_config(train_config)



import warnings
from typing import List, Optional, Sequence, Tuple

import torch.utils.data

from fme.core.dataset.config import XarrayDataConfig
from fme.core.dataset.xarray import DatasetProperties, XarrayDataset, get_xarray_dataset

from .requirements import DataRequirements


def get_datasets(
    dataset_configs: Sequence[XarrayDataConfig],
    requirements: DataRequirements,
    strict: bool = True,
) -> Tuple[List[XarrayDataset], DatasetProperties]:
    datasets = []
    properties: Optional[DatasetProperties] = None
    for config in dataset_configs:
        dataset, new_properties = get_xarray_dataset(config, requirements)
        datasets.append(dataset)
        if properties is None:
            properties = new_properties
        elif not strict:
            try:
                properties.update(new_properties)
            except ValueError as e:
                warnings.warn(
                    f"Metadata for each ensemble member are not the same: {e}"
                )
        else:
            properties.update(new_properties)
    if properties is None:
        raise ValueError("At least one dataset must be provided.")

    return datasets, properties


def get_dataset(
    dataset_configs: Sequence[XarrayDataConfig],
    requirements: DataRequirements,
    strict: bool = True,
) -> Tuple[torch.utils.data.ConcatDataset[XarrayDataset], DatasetProperties]:
    datasets, properties = get_datasets(dataset_configs, requirements, strict=strict)
    ensemble = torch.utils.data.ConcatDataset(datasets)
    return ensemble, properties



import dataclasses
import datetime
import os
import pathlib
from typing import List

import dacite
import numpy as np
import pytest
import torch
import xarray as xr
import yaml

from fme.ace.aggregator.inference import InferenceEvaluatorAggregatorConfig
from fme.ace.data_loading.inference import (
    InferenceDataLoaderConfig,
    InferenceInitialConditionIndices,
)
from fme.ace.inference.data_writer import DataWriterConfig
from fme.ace.inference.data_writer.time_coarsen import TimeCoarsenConfig
from fme.ace.inference.evaluator import InferenceEvaluatorConfig, main
from fme.ace.registry import ModuleSelector
from fme.ace.stepper import SingleModuleStepperConfig, TrainOutput
from fme.ace.testing import DimSizes, FV3GFSData, MonthlyReferenceData
from fme.core import metrics
from fme.core.coordinates import DimSize, HybridSigmaPressureCoordinate
from fme.core.dataset.config import XarrayDataConfig
from fme.core.device import get_device
from fme.core.gridded_ops import LatLonOperations
from fme.core.logging_utils import LoggingConfig
from fme.core.normalizer import NormalizationConfig
from fme.core.ocean import Ocean, OceanConfig
from fme.core.testing import mock_wandb
from fme.core.typing_ import TensorDict, TensorMapping

from .derived_variables import compute_derived_quantities

DIR = pathlib.Path(__file__).parent
TIMESTEP = datetime.timedelta(hours=6)


class PlusOne(torch.nn.Module):
    def forward(self, x):
        return x + 1


def save_plus_one_stepper(
    path: pathlib.Path,
    in_names: List[str],
    out_names: List[str],
    mean: float,
    std: float,
    data_shape: List[int],
    timestep: datetime.timedelta = TIMESTEP,
    nz_interface: int = 7,
):
    all_names = list(set(in_names).union(out_names))
    config = SingleModuleStepperConfig(
        builder=ModuleSelector(type="prebuilt", config={"module": PlusOne()}),
        in_names=in_names,
        out_names=out_names,
        normalization=NormalizationConfig(
            means={name: mean for name in all_names},
            stds={name: std for name in all_names},
        ),
    )
    area = torch.ones(data_shape[-2:], device=get_device())
    vertical_coordinate = HybridSigmaPressureCoordinate(
        ak=torch.arange(nz_interface), bk=torch.arange(nz_interface)
    )
    stepper = config.get_stepper(
        img_shape=(data_shape[-2], data_shape[-1]),
        gridded_operations=LatLonOperations(area),
        vertical_coordinate=vertical_coordinate,
        timestep=timestep,
    )
    torch.save({"stepper": stepper.get_state()}, path)


def test_inference_backwards_compatibility(tmp_path: pathlib.Path):
    """
    Inference test using a serialized model from an earlier commit, to ensure
    earlier models can be used with the updated inference code.
    """
    in_names = ["var"]
    out_names = ["var"]
    stepper_path = DIR / "stepper_test_data"

    horizontal = [DimSize("grid_yt", 4), DimSize("grid_xt", 8)]
    dim_sizes = DimSizes(
        n_time=8,
        horizontal=horizontal,
        nz_interface=2,
    )
    std = 1.0
    if not stepper_path.exists():
        # this helps us to re-generate data if the stepper changes
        # to re-generate, just delete the data and run the test (it will fail)
        save_plus_one_stepper(
            stepper_path,
            in_names,
            out_names,
            mean=0.0,
            std=std,
            data_shape=dim_sizes.shape_nd,
        )
        assert False, "stepper_test_data did not exist, it has been created"
    use_prediction_data = False
    n_forward_steps = 2
    inference_helper(
        tmp_path,
        in_names,
        out_names,
        use_prediction_data,
        dim_sizes,
        n_forward_steps,
        stepper_path,
        timestep=TIMESTEP,
    )


@pytest.mark.parametrize(
    "use_prediction_data, n_forward_steps",
    [(True, 2), (False, int(30 / 20 * 36)), (False, 1)],
)
def test_inference_plus_one_model(
    tmp_path: pathlib.Path, use_prediction_data: bool, n_forward_steps: int
):
    in_names = ["var"]
    out_names = ["var"]
    stepper_path = tmp_path / "stepper"

    horizontal = [DimSize("grid_yt", 16), DimSize("grid_xt", 32)]
    dim_sizes = DimSizes(
        n_time=n_forward_steps + 1,
        horizontal=horizontal,
        nz_interface=4,
    )
    if use_prediction_data:
        # use std of 10 so the stepper would have errors at the plus-one problem
        std = 10.0
    else:
        std = 1.0
    save_plus_one_stepper(
        stepper_path,
        in_names,
        out_names,
        mean=0.0,
        std=std,
        data_shape=dim_sizes.shape_nd,
        timestep=datetime.timedelta(days=20),
    )
    inference_helper(
        tmp_path,
        in_names,
        out_names,
        use_prediction_data,
        dim_sizes,
        n_forward_steps,
        stepper_path,
        save_monthly_files=False,  # requires timestep == 6h
        timestep=datetime.timedelta(days=20),
    )


def inference_helper(
    tmp_path,
    in_names,
    out_names,
    use_prediction_data,
    dim_sizes: DimSizes,
    n_forward_steps,
    stepper_path,
    timestep: datetime.timedelta,
    save_monthly_files: bool = True,
    derived_names: List[str] = [],
):
    time_varying_values = [float(i) for i in range(dim_sizes.n_time)]
    all_names = list(set(in_names).union(out_names))
    forcing_names = list(set(in_names).difference(out_names))
    data = FV3GFSData(
        path=tmp_path,
        names=all_names,
        dim_sizes=dim_sizes,
        time_varying_values=time_varying_values,
        timestep_days=timestep.total_seconds() / 86400,
        save_vertical_coordinate=False,
    )
    if use_prediction_data:
        prediction_data = data.inference_data_loader_config
    else:
        prediction_data = None

    if dim_sizes.n_time > 365 * 4:
        monthly_reference_filename = str(
            MonthlyReferenceData(
                path=pathlib.Path(tmp_path),
                names=all_names,
                dim_sizes=DimSizes(
                    n_time=48,
                    horizontal=dim_sizes.horizontal,
                    nz_interface=1,
                ),
                n_ensemble=3,
            ).data_filename
        )
    else:
        monthly_reference_filename = None
    config = InferenceEvaluatorConfig(
        experiment_dir=str(tmp_path),
        n_forward_steps=n_forward_steps,
        checkpoint_path=str(stepper_path),
        logging=LoggingConfig(
            log_to_screen=True,
            log_to_file=False,
            log_to_wandb=True,
        ),
        loader=data.inference_data_loader_config,
        prediction_loader=prediction_data,
        aggregator=InferenceEvaluatorAggregatorConfig(
            monthly_reference_data=monthly_reference_filename, log_video=True
        ),
        data_writer=DataWriterConfig(
            save_prediction_files=True,
            log_extended_video_netcdfs=True,
            save_histogram_files=True,
            save_monthly_files=save_monthly_files,
        ),
        forward_steps_in_memory=1,
    )
    config_filename = tmp_path / "config.yaml"
    with open(config_filename, "w") as f:
        yaml.dump(dataclasses.asdict(config), f)

    with mock_wandb() as wandb:
        wandb.configure(log_to_wandb=True)
        main(
            yaml_config=str(config_filename),
        )
        wandb_logs = wandb.get_logs()

    all_out_names = out_names + derived_names

    n_ic_timesteps = 1
    summary_log_step = 1
    assert len(wandb_logs) == n_ic_timesteps + config.n_forward_steps + summary_log_step
    for i in range(n_ic_timesteps + config.n_forward_steps):
        log = wandb_logs[i]
        for var in all_out_names:
            if i == 0 and var not in in_names:
                assert f"inference/mean/weighted_rmse/{var}" not in log
            else:
                # if these are off by something like 90% then probably the stepper
                # is being used instead of the prediction_data
                assert log[f"inference/mean/weighted_rmse/{var}"] == 0.0
                assert log[f"inference/mean/weighted_bias/{var}"] == 0.0

    var = list(set(in_names).difference(forcing_names))[0]

    if not use_prediction_data:
        initial_condition_ds = xr.open_dataset(
            tmp_path / "initial_condition.nc", decode_timedelta=False
        )
        for dim_name in ["lat", "lon"]:
            assert dim_name in initial_condition_ds.dims
            assert dim_name in initial_condition_ds.data_vars[var].dims
            assert dim_name in initial_condition_ds.coords

    prediction_ds = xr.open_dataset(
        tmp_path / "autoregressive_predictions.nc",
        decode_timedelta=False,
        decode_times=False,
    )
    assert len(prediction_ds["time"]) == config.n_forward_steps
    for i in range(config.n_forward_steps - 1):
        np.testing.assert_allclose(
            prediction_ds[var].isel(time=i).values + 1,
            prediction_ds[var].isel(time=i + 1).values,
        )
        assert not np.any(np.isnan(prediction_ds[var].isel(time=i + 1).values))

    assert "lat" in prediction_ds.coords
    assert "lon" in prediction_ds.coords

    if use_prediction_data:
        assert not os.path.exists(tmp_path / "restart.nc")
        assert not os.path.exists(tmp_path / "initial_condition.nc")
    else:
        restart_ds = xr.open_dataset(
            tmp_path / "restart.nc", decode_timedelta=False, decode_times=False
        )
        np.testing.assert_allclose(
            prediction_ds[var].isel(time=-1).values,
            restart_ds[var].values,
        )

        ic_ds = xr.open_dataset(
            tmp_path / "initial_condition.nc",
            decode_timedelta=False,
            decode_times=False,
        )
        np.testing.assert_allclose(ic_ds[var].values, 0.0)

    metric_ds = xr.open_dataset(tmp_path / "reduced_autoregressive_predictions.nc")
    assert var in metric_ds.data_vars
    assert metric_ds.data_vars[var].attrs["units"] == "m"
    assert metric_ds.data_vars[var].attrs["long_name"] == f"ensemble mean of {var}"
    assert f"rmse_{var}" in metric_ds.data_vars
    assert metric_ds.data_vars[f"rmse_{var}"].attrs["units"] == "m"
    assert (
        metric_ds.data_vars[f"rmse_{var}"].attrs["long_name"]
        == f"root mean squared error of {var}"
    )
    assert f"bias_{var}" in metric_ds.data_vars
    assert metric_ds.data_vars[f"bias_{var}"].attrs["units"] == "m"
    assert f"min_err_{var}" in metric_ds.data_vars
    assert metric_ds.data_vars[f"min_err_{var}"].attrs["units"] == "m"
    assert f"max_err_{var}" in metric_ds.data_vars
    assert metric_ds.data_vars[f"max_err_{var}"].attrs["units"] == "m"
    assert f"gen_var_{var}" in metric_ds.data_vars
    assert metric_ds.data_vars[f"gen_var_{var}"].attrs["units"] == ""
    assert (
        metric_ds.data_vars[f"gen_var_{var}"].attrs["long_name"]
        == f"prediction variance of {var} as fraction of target variance"
    )
    assert "lat" in metric_ds.coords
    assert "lon" in metric_ds.coords

    time_mean_diagnostics = xr.open_dataset(tmp_path / "time_mean_diagnostics.nc")
    actual_var_names = sorted([str(k) for k in time_mean_diagnostics.keys()])
    assert len(actual_var_names) == 2 * len(all_out_names)
    assert f"bias_map-{var}" in actual_var_names
    assert time_mean_diagnostics.data_vars[f"bias_map-{var}"].attrs["units"] == "m"
    assert f"gen_map-{var}" in actual_var_names
    assert time_mean_diagnostics.data_vars[f"gen_map-{var}"].attrs["units"] == "m"
    assert len(time_mean_diagnostics.coords) == 2
    assert "lat" in time_mean_diagnostics.coords
    assert "lon" in time_mean_diagnostics.coords

    zonal_mean_diagnostics = xr.open_dataset(tmp_path / "zonal_mean_diagnostics.nc")
    actual_var_names = sorted([str(k) for k in zonal_mean_diagnostics.keys()])
    assert len(actual_var_names) == 2 * len(all_out_names)
    assert f"error-{var}" in actual_var_names
    assert zonal_mean_diagnostics.data_vars[f"error-{var}"].attrs["units"] == "m"
    assert f"gen-{var}" in actual_var_names
    assert zonal_mean_diagnostics.data_vars[f"gen-{var}"].attrs["units"] == ""
    assert len(zonal_mean_diagnostics.coords) == 1
    assert "lat" in zonal_mean_diagnostics.coords

    for source in ["target", "prediction"]:
        histograms = xr.open_dataset(tmp_path / f"histograms_{source}.nc")
        actual_var_names = sorted([str(k) for k in histograms.keys()])
        # NOTE: target histograms include forcing variables
        n_vars = (
            len(all_out_names)
            if source == "prediction"
            else len(all_out_names) + len(forcing_names)
        )
        assert len(actual_var_names) == 2 * n_vars
        assert var in actual_var_names
        assert histograms.data_vars[var].attrs["units"] == "count"
        assert f"{var}_bin_edges" in actual_var_names
        assert histograms.data_vars[f"{var}_bin_edges"].attrs["units"] == "m"
        var_counts_per_timestep = histograms[var].sum(dim=["bin"])
        same_count_each_timestep = np.all(
            var_counts_per_timestep.values == var_counts_per_timestep.values[0]
        )
        assert same_count_each_timestep
    if monthly_reference_filename is not None:
        assert f"inference/annual/{var}" in wandb_logs[-1]
        assert f"inference/annual/r2_gen_{var}" in wandb_logs[-1]
        assert f"inference/annual/r2_target_{var}" in wandb_logs[-1]
    assert "inference/total_steps_per_second" in wandb_logs[-1]


@pytest.mark.parametrize(
    "n_forward_steps,forward_steps_in_memory", [(10, 2), (10, 10), (15, 4)]
)
def test_inference_writer_boundaries(
    tmp_path: pathlib.Path, n_forward_steps: int, forward_steps_in_memory: int
):
    """Test that data at initial condition boundaires"""
    in_names = ["var"]
    out_names = ["var"]
    all_names = list(set(in_names).union(out_names))
    stepper_path = tmp_path / "stepper"

    horizontal = [DimSize("grid_yt", 4), DimSize("grid_xt", 8)]

    dim_sizes = DimSizes(
        n_time=n_forward_steps + 1,
        horizontal=horizontal,
        nz_interface=4,
    )
    save_plus_one_stepper(
        stepper_path,
        in_names,
        out_names,
        mean=0.0,
        std=1.0,
        data_shape=dim_sizes.shape_nd,
    )
    data = FV3GFSData(
        path=tmp_path,
        names=all_names,
        dim_sizes=dim_sizes,
        timestep_days=TIMESTEP.total_seconds() / 86400,
    )
    config = InferenceEvaluatorConfig(
        experiment_dir=str(tmp_path),
        n_forward_steps=n_forward_steps,
        checkpoint_path=str(stepper_path),
        logging=LoggingConfig(log_to_screen=True, log_to_file=False, log_to_wandb=True),
        loader=data.inference_data_loader_config,
        data_writer=DataWriterConfig(
            save_prediction_files=True,
            time_coarsen=None,
        ),
        forward_steps_in_memory=forward_steps_in_memory,
    )
    config_filename = tmp_path / "config.yaml"
    with open(config_filename, "w") as f:
        yaml.dump(dataclasses.asdict(config), f)
    with mock_wandb() as wandb:
        wandb.configure(log_to_wandb=True)
        main(
            yaml_config=str(config_filename),
        )
        inference_logs = wandb.get_logs()
    n_ic_timesteps = 1
    summary_log_step = 1
    assert (
        len(inference_logs)
        == n_ic_timesteps + config.n_forward_steps + summary_log_step
    )

    prediction_ds = xr.open_dataset(
        tmp_path / "autoregressive_predictions.nc", decode_timedelta=False
    )
    target_ds = xr.open_dataset(
        tmp_path / "autoregressive_target.nc", decode_timedelta=False
    )
    # data writers do not include initial condition
    assert len(prediction_ds["time"]) == n_forward_steps
    assert not np.any(np.isnan(prediction_ds["var"].values))

    gen = prediction_ds["var"]
    tar = target_ds["var"]
    gen_time_mean = torch.from_numpy(gen.mean(dim="time").values)
    tar_time_mean = torch.from_numpy(tar.mean(dim="time").values)
    area_weights = metrics.spherical_area_weights(
        tar["lat"].values, num_lon=len(tar["lon"])
    )
    # check time mean metrics
    tol = 1e-4  # relative tolerance
    assert metrics.root_mean_squared_error(
        tar_time_mean, gen_time_mean, area_weights
    ).item() == pytest.approx(
        inference_logs[-1]["inference/time_mean/rmse/var"], rel=tol
    )
    assert metrics.weighted_mean_bias(
        tar_time_mean, gen_time_mean, area_weights
    ).item() == pytest.approx(
        inference_logs[-1]["inference/time_mean/bias/var"], rel=tol
    )

    prediction_ds = prediction_ds.isel(sample=0)
    target_ds = target_ds.isel(sample=0)
    ds = xr.open_dataset(data.data_filename)

    for i in range(0, n_forward_steps):
        # metrics logs includes IC while saved data does not
        log = inference_logs[i + n_ic_timesteps]
        # metric steps should match lead times
        assert log["inference/mean/forecast_step"] == i + n_ic_timesteps
        gen_i = torch.from_numpy(gen.isel(time=i).values)
        tar_i = torch.from_numpy(tar.isel(time=i).values)
        # check that manually computed metrics match logged metrics
        assert metrics.root_mean_squared_error(
            tar_i, gen_i, area_weights, dim=(-2, -1)
        ).item() == pytest.approx(log["inference/mean/weighted_rmse/var"], rel=tol)
        assert metrics.weighted_mean_bias(
            tar_i, gen_i, area_weights, dim=(-2, -1)
        ).item() == pytest.approx(log["inference/mean/weighted_bias/var"], rel=tol)
        assert metrics.gradient_magnitude_percent_diff(
            tar_i, gen_i, area_weights, dim=(-2, -1)
        ).item() == pytest.approx(
            log["inference/mean/weighted_grad_mag_percent_diff/var"], rel=tol
        )
        assert metrics.weighted_mean(
            gen_i, area_weights, dim=(-2, -1)
        ).item() == pytest.approx(log["inference/mean/weighted_mean_gen/var"], rel=tol)

        # the target obs should be the same as the validation data obs
        # ds is original data which includes IC, target_ds does not
        np.testing.assert_allclose(
            target_ds["var"].isel(time=i).values,
            ds["var"].isel(time=i + 1).values,
        )
        if i > 0:
            lead_da = prediction_ds["var"].isel(time=i)
            # predictions should be previous condition + 1
            np.testing.assert_allclose(
                lead_da.values,
                prediction_ds["var"].isel(time=i - 1).values + 1,
            )
            # prediction and target should not have entirely the same values at
            # any lead > 0
            assert not np.allclose(
                lead_da.values,
                target_ds["var"].isel(time=i).values,
            )


def test_inference_data_time_coarsening(tmp_path: pathlib.Path):
    forward_steps_in_memory = 4
    coarsen_factor = 2
    in_names = ["var"]
    out_names = ["var"]
    all_names = list(set(in_names).union(out_names))
    stepper_path = tmp_path / "stepper"

    horizontal = [DimSize("grid_yt", 16), DimSize("grid_xt", 32)]

    dim_sizes = DimSizes(
        n_time=9,
        horizontal=horizontal,
        nz_interface=4,
    )
    save_plus_one_stepper(
        stepper_path,
        in_names,
        out_names,
        mean=0.0,
        std=1.0,
        data_shape=dim_sizes.shape_nd,
    )
    data = FV3GFSData(
        path=tmp_path,
        names=all_names,
        dim_sizes=dim_sizes,
        timestep_days=TIMESTEP.total_seconds() / 86400,
    )
    config = InferenceEvaluatorConfig(
        experiment_dir=str(tmp_path),
        n_forward_steps=8,
        forward_steps_in_memory=forward_steps_in_memory,
        checkpoint_path=str(stepper_path),
        logging=LoggingConfig(
            log_to_screen=True,
            log_to_file=False,
            log_to_wandb=False,
        ),
        loader=data.inference_data_loader_config,
        data_writer=DataWriterConfig(
            save_prediction_files=True,
            log_extended_video_netcdfs=True,
            time_coarsen=TimeCoarsenConfig(coarsen_factor=coarsen_factor),
            save_histogram_files=True,
        ),
    )
    config_filename = tmp_path / "config.yaml"
    with open(config_filename, "w") as f:
        yaml.dump(dataclasses.asdict(config), f)
    main(yaml_config=str(config_filename))
    # check that the outputs all have the intended time dimension size
    prediction_ds = xr.open_dataset(
        tmp_path / "autoregressive_predictions.nc", decode_timedelta=False
    )
    n_coarsened_timesteps = config.n_forward_steps // coarsen_factor
    assert (
        len(prediction_ds["time"]) == n_coarsened_timesteps
    ), "raw predictions time dimension size"
    metric_ds = xr.open_dataset(tmp_path / "reduced_autoregressive_predictions.nc")
    assert (
        metric_ds.sizes["timestep"] == n_coarsened_timesteps
    ), "reduced predictions time dimension size"
    for source in ["target", "prediction"]:
        histograms = xr.open_dataset(tmp_path / f"histograms_{source}.nc")
        assert (
            histograms.sizes["time"] == n_coarsened_timesteps
        ), "histograms time dimension size"


@pytest.mark.parametrize("has_required_fields", [True, False])
def test_compute_derived_quantities(has_required_fields):
    """Checks that tensors are added to the data dictionary appropriately."""
    n_sample, n_time, nx, ny, nz = 2, 3, 4, 5, 6

    def _make_data():
        vars = ["a"]
        if has_required_fields:
            additional_fields = [
                "specific_total_water_{}".format(i) for i in range(nz)
            ] + ["PRESsfc"]
            vars += additional_fields
        return {
            var: torch.randn(n_sample, n_time, nx, ny, device=get_device())
            for var in vars
        }

    vertical_coordinate = HybridSigmaPressureCoordinate(
        ak=torch.linspace(0, 1, nz + 1, device=get_device()),
        bk=torch.linspace(0, 1, nz + 1, device=get_device()),
    )

    def derive_func(data: TensorMapping, forcing_data: TensorMapping) -> TensorDict:
        updated = compute_derived_quantities(
            dict(data),
            vertical_coordinate=vertical_coordinate,
            timestep=TIMESTEP,
            forcing_data=dict(forcing_data),
        )
        return updated

    metrics = {"loss": 42.0}
    fake_data = {k: _make_data() for k in ("gen_data", "target_data")}
    stepped = TrainOutput(
        metrics,
        fake_data["gen_data"],
        fake_data["target_data"],
        time=xr.DataArray(np.zeros((n_sample, n_time)), dims=["sample", "time"]),
        normalize=lambda x: x,
        derive_func=derive_func,
    )

    derived_stepped = stepped.compute_derived_variables()

    dry_air_name = "surface_pressure_due_to_dry_air"
    water_path_name = "total_water_path"
    existence_check = (
        dry_air_name in derived_stepped.gen_data
        and dry_air_name in derived_stepped.target_data
        and water_path_name in derived_stepped.gen_data
        and water_path_name in derived_stepped.target_data
    )

    if has_required_fields:
        assert existence_check
        fields = (
            derived_stepped.gen_data[dry_air_name],
            derived_stepped.target_data[dry_air_name],
            derived_stepped.gen_data["a"],
            derived_stepped.target_data["a"],
        )
        for f in fields:
            assert f.shape == (n_sample, n_time, nx, ny)
    else:
        assert not existence_check


def test_derived_metrics_run_without_errors(tmp_path: pathlib.Path):
    """Checks that derived metrics are computed during inferece without errors."""

    n_forward_steps = 2

    in_names = ["var", "PRESsfc", "specific_total_water_0", "specific_total_water_1"]
    out_names = ["var", "PRESsfc", "specific_total_water_0", "specific_total_water_1"]
    all_names = list(set(in_names).union(out_names))
    stepper_path = tmp_path / "stepper"

    horizontal = [DimSize("grid_yt", 16), DimSize("grid_xt", 32)]

    dim_sizes = DimSizes(
        n_time=n_forward_steps + 1,
        horizontal=horizontal,
        nz_interface=3,
    )
    save_plus_one_stepper(
        stepper_path,
        in_names,
        out_names,
        mean=0.0,
        std=1.0,
        data_shape=dim_sizes.shape_nd,
        nz_interface=dim_sizes.nz_interface,
    )
    time_varying_values = [float(i) for i in range(dim_sizes.n_time)]
    data = FV3GFSData(
        path=tmp_path,
        names=all_names,
        dim_sizes=dim_sizes,
        time_varying_values=time_varying_values,
        timestep_days=TIMESTEP.total_seconds() / 86400,
        num_data_workers=2,
    )
    config = InferenceEvaluatorConfig(
        experiment_dir=str(tmp_path),
        n_forward_steps=n_forward_steps,
        checkpoint_path=str(stepper_path),
        logging=LoggingConfig(
            log_to_screen=True,
            log_to_file=False,
            log_to_wandb=True,
        ),
        loader=data.inference_data_loader_config,
        prediction_loader=None,
        data_writer=DataWriterConfig(save_prediction_files=True),
        forward_steps_in_memory=1,
    )
    config_filename = tmp_path / "config.yaml"
    with open(config_filename, "w") as f:
        yaml.dump(dataclasses.asdict(config), f)

    with mock_wandb() as wandb:
        wandb.configure(log_to_wandb=True)
        main(yaml_config=str(config_filename))
        inference_logs = wandb.get_logs()

    # derived variables should not have normalized metrics reported
    assert "inference/mean_norm/weighted_rmse/total_water_path" not in inference_logs[0]
    assert "inference/time_mean_norm/rmse/total_water_path" not in inference_logs[-1]


@pytest.mark.parametrize(
    "time_coarsen,n_forward_steps,forward_steps_in_memory",
    [
        pytest.param(3, 12, 4, id="not_multiple_of_forward_steps_in_memory"),
        pytest.param(-1, 12, 4, id="invalid_time_coarsen"),
        pytest.param(2, 5, 4, id="not_multiple_of_n_forward_steps"),
    ],
)
def test_inference_config_raises_incompatible_timesteps(
    time_coarsen, n_forward_steps, forward_steps_in_memory
):
    base_config_dict = dict(
        experiment_dir="./some_dir",
        n_forward_steps=n_forward_steps,
        checkpoint_path="./some_dir",
        logging=LoggingConfig(),
        loader=InferenceDataLoaderConfig(
            XarrayDataConfig(
                data_path="./some_data",
            ),
            start_indices=InferenceInitialConditionIndices(
                first=0, n_initial_conditions=1, interval=1
            ),
        ),
    )
    base_config_dict["forward_steps_in_memory"] = forward_steps_in_memory
    base_config_dict["data_writer"] = {"time_coarsen": {"coarsen_factor": time_coarsen}}
    with pytest.raises(ValueError):
        dacite.from_dict(
            data_class=InferenceEvaluatorConfig,
            data=base_config_dict,
            config=dacite.Config(strict=True),
        )


def test_inference_ocean_override(tmp_path: pathlib.Path):
    """Test that data at initial condition boundaires"""
    in_names = ["var"]
    out_names = ["var"]
    all_names = list(set(in_names).union(out_names))
    stepper_path = tmp_path / "stepper"
    n_forward_steps = 8

    horizontal = [DimSize("grid_yt", 4), DimSize("grid_xt", 8)]
    dim_sizes = DimSizes(
        n_time=n_forward_steps + 1,
        horizontal=horizontal,
        nz_interface=4,
    )
    save_plus_one_stepper(
        stepper_path,
        in_names,
        out_names,
        mean=0.0,
        std=1.0,
        data_shape=dim_sizes.shape_nd,
    )
    data = FV3GFSData(
        path=tmp_path,
        names=all_names,
        dim_sizes=dim_sizes,
        timestep_days=TIMESTEP.total_seconds() / 86400,
    )
    ocean_override = OceanConfig(
        surface_temperature_name="var",
        ocean_fraction_name="override_ocean_fraction",
    )

    config = InferenceEvaluatorConfig(
        experiment_dir=str(tmp_path),
        n_forward_steps=n_forward_steps,
        checkpoint_path=str(stepper_path),
        logging=LoggingConfig(log_to_screen=True, log_to_file=False, log_to_wandb=True),
        loader=data.inference_data_loader_config,
        data_writer=DataWriterConfig(save_prediction_files=True, time_coarsen=None),
        forward_steps_in_memory=4,
        ocean=ocean_override,
    )
    stepper = config.load_stepper()
    assert isinstance(stepper.ocean, Ocean)
    assert (
        stepper.ocean.surface_temperature_name
        == ocean_override.surface_temperature_name
    )
    assert stepper.ocean.ocean_fraction_name == ocean_override.ocean_fraction_name

    stepper_config = config.load_stepper_config()
    assert stepper_config.ocean == ocean_override


def test_inference_timestep_mismatch_error(tmp_path: pathlib.Path):
    """Test that inference with a model trained with a different timestep than
    the forcing data raises an error.
    """
    in_names = ["var"]
    out_names = ["var"]
    stepper_path = tmp_path / "stepper_test_data"

    horizontal = [DimSize("grid_yt", 4), DimSize("grid_xt", 8)]
    dim_sizes = DimSizes(n_time=8, horizontal=horizontal, nz_interface=2)
    std = 1.0
    save_plus_one_stepper(
        stepper_path,
        in_names,
        out_names,
        mean=0.0,
        std=std,
        data_shape=dim_sizes.shape_nd,
        timestep=TIMESTEP,
    )
    use_prediction_data = False
    n_forward_steps = 2
    with pytest.raises(ValueError, match="Timestep of the loaded stepper"):
        inference_helper(
            tmp_path,
            in_names,
            out_names,
            use_prediction_data,
            dim_sizes,
            n_forward_steps,
            stepper_path,
            timestep=datetime.timedelta(days=20),
        )


def test_inference_includes_diagnostics(tmp_path: pathlib.Path):
    """Test that diagnostics are included in evaluator metrics and outputs."""
    # NOTE: size of in_names and out_names has to be the same here or the
    # PlusOne outputs won't have the right shape
    in_names = ["prog", "forcing_var", "DSWRFtoa"]
    out_names = ["prog", "ULWRFtoa", "USWRFtoa"]
    stepper_path = tmp_path / "stepper"
    horizontal = [DimSize("grid_yt", 16), DimSize("grid_xt", 32)]
    use_prediction_data = False
    n_forward_steps = 2
    dim_sizes = DimSizes(
        n_time=n_forward_steps + 1,
        horizontal=horizontal,
        nz_interface=4,
    )
    save_plus_one_stepper(
        stepper_path,
        in_names,
        out_names,
        mean=0.0,
        std=1.0,
        data_shape=dim_sizes.shape_nd,
        timestep=datetime.timedelta(days=20),
    )
    inference_helper(
        tmp_path,
        in_names,
        out_names,
        use_prediction_data,
        dim_sizes,
        n_forward_steps,
        stepper_path,
        save_monthly_files=False,  # requires timestep == 6h
        timestep=datetime.timedelta(days=20),
        derived_names=["net_energy_flux_toa_into_atmosphere"],
    )
    ds = xr.open_dataset(
        tmp_path / "autoregressive_predictions.nc",
        decode_timedelta=False,
        decode_times=False,
    )
    # prognostic in
    assert "prog" in ds
    # diags in
    assert "ULWRFtoa" in ds
    assert "USWRFtoa" in ds
    # derived in
    assert "net_energy_flux_toa_into_atmosphere" in ds
    # forcings not in
    assert "DSWRFtoa" not in ds
    assert "forcing_var" not in ds
    # assert only prognostic variables are in initial condition and restart files
    for filename in ["initial_condition.nc", "restart.nc"]:
        ds = xr.open_dataset(tmp_path / filename)
        assert "USWRFtoa" not in ds
        assert "forcing_var" not in ds
        assert "prog" in ds
